#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from os.path import basename, normpath
from typing import Any, Collection, Dict, List, Optional, Tuple

import pandas as pd

from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import \
    generate_read_csv_code
from mitosheet.preprocessing.preprocess_step_performer import \
    PreprocessStepPerformer
from mitosheet.step_performers.import_steps.simple_import import \
    read_csv_get_delimiter_and_encoding
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.transpiler.transpile_utils import get_str_param_name
from mitosheet.types import StepsManagerType
from mitosheet.utils import get_valid_dataframe_name


class ReadFilePathsPreprocessStepPerformer(PreprocessStepPerformer):
    """
    This preprocessor reads in any arguments that are
    strings, treats them as file paths, and attempts
    to read them in a dataframes.
    """

    @classmethod
    def preprocess_step_version(cls) -> int:
        return 1

    @classmethod
    def preprocess_step_type(cls) -> str:
        return 'read_file_paths'

    @classmethod
    def execute(cls, args: Collection[Any]) -> Tuple[List[Any], Optional[List[str]], Optional[Dict[str, Any]]]:
        df_args: List[pd.DataFrame] = []
        delimeters: List[Optional[str]] = []
        encodings: List[Optional[str]] = []
        df_names: List[str] = []

        for arg_index, arg in enumerate(args):
            if isinstance(arg, pd.DataFrame):
                df_args.append(arg)
                delimeters.append(None)
                encodings.append(None)
                df_names.append('df' + str(arg_index + 1))
            elif isinstance(arg, str):
                # If it is a string, we try and read it in as a dataframe
                try:
                    # We use the simple import 
                    df, delimeter, encoding = read_csv_get_delimiter_and_encoding(arg)

                    df_args.append(
                        df
                    )

                    delimeters.append(delimeter)
                    encodings.append(encoding)
                    df_names.append(get_valid_dataframe_name(df_names, basename(normpath(arg))))
                except:
                    # If this pd.read_csv fails, then we report this error to the user
                    # as a failed mitosheet call
                    error_message = f'Invalid argument passed to sheet: {arg}. This path could not be read with a pd.read_csv call. Please pass in the parsed dataframe directly.'
                    log('mitosheet_sheet_call_failed', failed=True)
                    raise ValueError(error_message)
            else:
                error_message = f'Invalid argument passed to sheet: {arg}. Please pass all dataframes or paths to CSV files.'
                log('mitosheet_sheet_call_failed', {'error': error_message}, failed=True)
                raise ValueError(error_message)
                
        return df_args, df_names, {
            'delimeters': delimeters,
            'encodings': encodings,
            'df_names': df_names
        }

    @classmethod
    def transpile(cls, steps_manager: StepsManagerType, execution_data: Optional[Dict[str, Any]]) -> Tuple[List[str], List[str]]:
        """
        Transpiles the reading in of passed file paths to dataframe names, 
        with a simple pd.read_csv call.
        """
        code = []
        
        delimeters = execution_data['delimeters'] if execution_data is not None else []
        encodings = execution_data['encodings'] if execution_data is not None else []
        df_names = execution_data['df_names'] if execution_data is not None else []

        for arg_index, arg in enumerate(steps_manager.original_args):
            if isinstance(arg, str):
                df_name = df_names[arg_index]

                # Make sure to compile the path as a variable if the user is creating a function
                file_name = arg if not steps_manager.code_options['as_function'] else get_str_param_name(steps_manager, arg_index)
                read_csv_code = generate_read_csv_code(
                    file_name, df_name, delimeters[arg_index], encodings[arg_index], None, None, None, file_name_is_variable=steps_manager.code_options['as_function']
                )

                code.append(
                    read_csv_code
                )

        if len(code) > 0:
            code.insert(0, '# Read in filepaths as dataframes')
                
        return code, ['import pandas as pd'] if len(code) > 0 else []


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import os
from mitosheet.telemetry.telemetry_utils import log

IPYTHON_STARTUP_FOLDER = os.path.join(os.path.expanduser("~"), '.ipython', 'profile_default', 'startup')
IMPORT_MITOSHEET_FILE_PATH = os.path.join(IPYTHON_STARTUP_FOLDER, 'import_mitosheet.py')

# The files in the startup folder, including the file we're creating here
# are executed when the IPython kernel is created. This occurs when starting a
# Juptyer Lab server.

IMPORT_MITOSHEET_FILE_CONTENTS = """
try:
\timport mitosheet
\timport pandas as pd
\timport uuid

\tdef add_mito_button_to_df_output(obj):
\t\tbutton_uuid = uuid.uuid4()
\t\ttry:
\t\t\tmax_rows = pd.get_option('display.min_rows') # NOTE: this is due to https://github.com/pandas-dev/pandas/issues/44304
\t\t\tmax_cols = pd.get_option('display.max_columns')
\t\texcept:
\t\t\tmax_rows = 10
\t\t\tmax_cols = 20
\t\treturn f'<div><div id={button_uuid} style="display:none; background-color:#9D6CFF; color:white; width:200px; height:30px; padding-left:5px; border-radius:4px; flex-direction:row; justify-content:space-around; align-items:center;" onmouseover="this.style.backgroundColor=\\'#BA9BF8\\'" onmouseout="this.style.backgroundColor=\\'#9D6CFF\\'" onclick="window.commands?.execute(\\'create-mitosheet-from-dataframe-output\\');">See Full Dataframe in Mito</div> <script> if (window.commands?.hasCommand(\\'create-mitosheet-from-dataframe-output\\')) document.getElementById(\\'{button_uuid}\\').style.display = \\'flex\\' </script> {obj.to_html(max_rows=max_rows, max_cols=max_cols)}</div>'

\thtml_formatter = get_ipython().display_formatter.formatters['text/html']
\thtml_formatter.for_type(pd.DataFrame, add_mito_button_to_df_output)
except:
\tprint('Unable to automatically import mitosheet')
"""

def create_startup_file() -> None:
    """
    Adds a file import_mitosheet.py to the user's IPython startup folder 
    so that Mito is automatically imported everytime they start an IPython kernel. 

    This allows us to display a button in the user's pandas dataframe output that 
    lets them open the dataframe in Mito.
    """

    # Create the startup folder if it does not exist
    if not os.path.exists(IPYTHON_STARTUP_FOLDER):
        os.makedirs(IPYTHON_STARTUP_FOLDER)
        
    # Create the import mitosheet file
    with open(IMPORT_MITOSHEET_FILE_PATH, 'w+') as f:
        # And write the default object
        f.write(IMPORT_MITOSHEET_FILE_CONTENTS)

    log('created_df_button_startup_file')


def remove_startup_file() -> None:
    """
    Removes the import_mitosheet.py from the user's IPython startup folder
    so that the View in Mito button is not automatically displayed in 
    every dataframe output
    """
    os.remove(IMPORT_MITOSHEET_FILE_PATH)
    log('removed_df_button_startup_file')




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import os
from typing import Any, Dict, List, Optional, Tuple

import requests  # type: ignore

from mitosheet.ai.prompt import PROMPT_VERSION, get_prompt
from mitosheet.types import Selection, StepsManagerType
from mitosheet.user.db import get_user_field, set_user_field
from mitosheet.user.schemas import (UJ_AI_MITO_API_NUM_USAGES,
                                    UJ_STATIC_USER_ID, UJ_USER_EMAIL)
from mitosheet.user.utils import is_pro

OPEN_AI_URL = 'https://api.openai.com/v1/chat/completions'
MITO_AI_URL = 'https://ogtzairktg.execute-api.us-east-1.amazonaws.com/Prod/completions/'

OPEN_SOURCE_AI_COMPLETIONS_LIMIT = 100

def _get_ai_completion_data(prompt: str) -> Dict[str, Any]:
        return {
                "model": "gpt-3.5-turbo",
                "messages": [{"role": "user", "content": prompt}],
                "max_tokens": 7,
                "temperature": .2,
                'max_tokens': 200,
                'stop': ['```']
        }

__user_email = None
__user_id = None
__num_usages = None

def _get_ai_completion_from_mito_server(user_input: str, prompt: str) -> Dict[str, Any]:
        global __user_email, __user_id, __num_usages

        if __user_email is None:
                __user_email = get_user_field(UJ_USER_EMAIL)
        if __user_id is None:
                __user_id = get_user_field(UJ_STATIC_USER_ID)
        if __num_usages is None:
                __num_usages = get_user_field(UJ_AI_MITO_API_NUM_USAGES)


        if __num_usages is None:
                __num_usages = 0

        pro = is_pro()

        if not pro and __num_usages >= OPEN_SOURCE_AI_COMPLETIONS_LIMIT:
                return {
                        'error': f'You have used Mito AI {OPEN_SOURCE_AI_COMPLETIONS_LIMIT} times.'
                }
                

        data = {
                'email': __user_email,
                'user_id': __user_id,
                'user_input': user_input,
                'data': _get_ai_completion_data(prompt)
        }

        headers = {
                'Content-Type': 'application/json',
        }

        try:
                res = requests.post(MITO_AI_URL, headers=headers, json=data)
        except:
                return {
                        'error': f'There was an error accessing the Mito AI API. This is likely due to internet connectivity problems or a firewall.'
                }
        

        if res.status_code == 200:
                __num_usages = __num_usages + 1
                set_user_field(UJ_AI_MITO_API_NUM_USAGES, __num_usages + 1)
                return {
                        'user_input': user_input,
                        'prompt_version': PROMPT_VERSION,
                        'prompt': prompt,
                        'completion': res.json()['completion'],
                }
        
        return {
                'error': f'There was an error accessing the MitoAI API. {res.json()["error"]}'
        }

def _get_ai_completion_from_open_ai_api_compatible_server(url: str, user_input: str, prompt: str) -> Dict[str, Any]:

        data = _get_ai_completion_data(prompt)
        headers = {
                'Content-Type': 'application/json',
        }

        try:
                res = requests.post(url, headers=headers, json=data)
        except:
                return {
                        'error': f'There was an error accessing the API at {url}. This is likely due to internet connectivity problems or a firewall.'
                }

        if res.status_code == 200:
                res_json = res.json()
                completion: str = res_json['choices'][0]['message']["content"]
                # We strip all blank lines from the generated code, if they are at the start or end
                completion = completion.strip()
                return {
                        'user_input': user_input,
                        'prompt_version': PROMPT_VERSION,
                        'prompt': prompt,
                        'completion': completion,
                }

        return {
                'error': f'There was an error accessing the API at {url}. {res.json()["error"]["message"]}'
        }



def get_ai_completion(params: Dict[str, Any], steps_manager: StepsManagerType) -> Dict[str, Any]:
        selection: Optional[Selection] = params.get('selection', None)
        user_input: str = params['user_input']
        previous_failed_completions: List[Tuple[str, str]] = params['previous_failed_completions']

        prompt = get_prompt(
                steps_manager.curr_step.final_defined_state.df_names,
                steps_manager.curr_step.final_defined_state.dfs,
                selection,
                user_input,
                previous_failed_completions,
        )

        OPENAI_API_KEY = os.environ.get('OPENAI_API_KEY')
        byo_url = steps_manager.mito_config.get_llm_url()

        if byo_url is not None:
                return _get_ai_completion_from_open_ai_api_compatible_server(byo_url, user_input, prompt)
        elif OPENAI_API_KEY is None:
                # If they don't have an Open AI key, we use the mito server to get a completion
                return _get_ai_completion_from_mito_server(user_input, prompt)


        data = _get_ai_completion_data(prompt)
        headers = {
                'Content-Type': 'application/json',
                'Authorization': f'Bearer {OPENAI_API_KEY}' 
        }

        try:
                res = requests.post(OPEN_AI_URL, headers=headers, json=data)
        except:
                return {
                        'error': f'There was an error accessing the OpenAI API. This is likely due to internet connectivity problems or a firewall.'
                }
        
        if res.status_code == 200:
                res_json = res.json()
                completion: str = res_json['choices'][0]['message']["content"]
                # We strip all blank lines from the generated code, if they are at the start or end
                completion = completion.strip()
                return {
                        'user_input': user_input,
                        'prompt_version': PROMPT_VERSION,
                        'prompt': prompt,
                        'completion': completion,
                }

        return {
                'error': f'There was an error accessing the OpenAI API. {res.json()["error"]["message"]}'
        }


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
For each column header, Mito creates a static `column_id` that
is used to identify that column header throughout the entire course
of the analysis. 

Anywhere where we refer to a column_header, we refer to it by its id, 
and that way if we rename/reorder the column headers, we can still 
access the specific column that we want. 

Notably, the ID of the column header is just a string representation
of the column header. For string column headers, they are just
themselves.
"""
import random
from typing import Any, Collection, Dict, List, Optional
import numpy as np

import pandas as pd

from mitosheet.errors import make_no_column_error
from mitosheet.types import ColumnHeader, ColumnID, MultiLevelColumnHeader


def get_column_header_display(column_header: ColumnHeader) -> str:
    """
    Turns the column header into a string that is the same as how
    the string is displayed on the front-end. 

    This is useful for, say, the parser, where it gets data from
    the front-end - where the users enters column headers as they 
    see them.

    NOTE: must match the equivalent function in src/utils/columnHeaders.tsx
    """
    if isinstance(column_header, bool):
        return 'true' if column_header else 'false'
    elif isinstance(column_header, tuple):
        return ', '.join([get_column_header_display(c) for c in column_header if c != ''])

    return str(column_header)


def get_column_header_ids(column_headers: List[ColumnHeader]) -> List[ColumnID]:
    return [
        get_column_header_id(column_header) for column_header in column_headers
    ]

def get_column_header_id(column_header: ColumnHeader, use_deprecated_id_algorithm: bool=False) -> ColumnID:
    """
    Returns the ID that represents this column header, which is
    used to refer to the column as long as it is used in the 
    analysis.

    NOTE: this should only be called the first time the column
    header is seen, and so is static throughout the analysis.

    If use_deprecated_id_algorithm is True, then we use the old 
    algorithm for generating the IDs. This is necessary for back
    wards compatibility.
    """
    if use_deprecated_id_algorithm:
        from mitosheet.step_performers.bulk_old_rename.deprecated_utils import \
            make_valid_header
        return make_valid_header(column_header)

    if isinstance(column_header, str):
        return column_header
    else:
        return str(column_header)

def try_make_new_header_valid_if_multi_index_headers(column_headers: List[ColumnHeader], new_column_header: ColumnHeader) -> ColumnHeader:
    """
    Helper function for when you're adding a column to a
    dataframe that has multi-index headers. 

    In this case, you need to add a new header in a tuple
    that is the length of the multi-index headers, so
    this function helps you make such a header
    """
    if (not isinstance(new_column_header, tuple) and not isinstance(new_column_header, list)) and len(column_headers) > 0:
        other_column_header = random.choice(list(column_headers))
        if isinstance(other_column_header, tuple):
            ending_list: MultiLevelColumnHeader = [new_column_header] + ['' for _ in range(len(other_column_header) - 1)]
            return tuple(ending_list)

    # If we don't need to change the column header, don't change it
    return new_column_header

    

class ColumnIDMap():
    """
    This utility class maintains a mapping from column_id -> column_header
    and the reverse, so that it is easy to move back and forth between the
    two of them.
    """

    def __init__(self, dfs: Collection[pd.DataFrame]):
        self.column_id_to_column_header: List[Dict[ColumnID, ColumnHeader]] = [dict() for _ in range(len(dfs))]
        self.column_header_to_column_id: List[Dict[ColumnHeader, ColumnID]] = [dict() for _ in range(len(dfs))]

        for sheet_index, df in enumerate(dfs):
            for column_header in df.keys():
                column_id = get_column_header_id(column_header)
                self.column_id_to_column_header[sheet_index][column_id] = column_header
                self.column_header_to_column_id[sheet_index][column_header] = column_id

    def set_column_header(self, sheet_index: int, column_id: ColumnID, column_header: ColumnHeader) -> None:
        """
        Sets a column id and column header to match to eachother. 

        Takes special care in the case that the existing dataframe is a multi-index
        headers, in which case even if the new column header is just a string, the
        the column header that ends up getting created in the dataframe will be a 
        tuple.
        """
        # Make sure the new column header we're setting is in fact valid for this dataframe
        # header format - as it may be a multi-index header
        column_header = try_make_new_header_valid_if_multi_index_headers(
            list(self.column_id_to_column_header[sheet_index].values()),
            column_header
        )

        old_column_header = self.column_id_to_column_header[sheet_index][column_id]
        del self.column_header_to_column_id[sheet_index][old_column_header]
        
        self.column_id_to_column_header[sheet_index][column_id] = column_header
        self.column_header_to_column_id[sheet_index][column_header] = column_id

    def add_df(self, df: pd.DataFrame, sheet_index: Optional[int]=None, use_deprecated_id_algorithm: bool=False) -> Dict[str, ColumnHeader]:
        """
        Adds all of the keys for the new dataframe to the column id 
        mappings.

        Returns a mapping from column ids -> column headers for this
        dataframe, which is then used to set other variables in the 
        state.
        """
        if sheet_index is None:
            self.column_id_to_column_header.append(dict())
            self.column_header_to_column_id.append(dict())
            sheet_index = len(self.column_header_to_column_id) - 1
        else:
            self.column_id_to_column_header[sheet_index] = dict()
            self.column_header_to_column_id[sheet_index] = dict()
        
        for column_header in df.keys():
            column_id = get_column_header_id(column_header, use_deprecated_id_algorithm=use_deprecated_id_algorithm)
            self.column_id_to_column_header[sheet_index][column_id] = column_header
            self.column_header_to_column_id[sheet_index][column_header] = column_id
        
        return self.column_id_to_column_header[sheet_index]

    def remove_df(self, sheet_index: int) -> None:
        """
        Deletes the tracking of this dataframe from the column
        id map.
        """
        self.column_id_to_column_header.pop(sheet_index)
        self.column_header_to_column_id.pop(sheet_index)
    
    def add_column_header(self, sheet_index: int, column_header: ColumnHeader) -> str:
        """
        NOTE: this should only be called when adding a column to the dataframe,
        and not when renaming a column, as it creates a new id for the column
        header!
        """
        # Make sure the new column header we're setting is in fact valid for this dataframe
        # header format - as it may be a multi-index header
        column_header = try_make_new_header_valid_if_multi_index_headers(
            list(self.column_id_to_column_header[sheet_index].values()),
            column_header
        )

        column_id = get_column_header_id(column_header)
        self.column_id_to_column_header[sheet_index][column_id] = column_header
        self.column_header_to_column_id[sheet_index][column_header] = column_id

        return column_id
    
    def delete_column_id(self, sheet_index: int, column_id: ColumnID) -> None:
        column_header = self.get_column_header_by_id(sheet_index, column_id)
        del self.column_id_to_column_header[sheet_index][column_id]
        self.column_header_to_column_id[sheet_index][column_header]

    # Note: The object returned by this function might not reflect the actual order of columns.
    # To access the list of columns in order, use the dataframe's columns.tolist() method. 
    def get_column_ids(self, sheet_index: int, column_headers: Optional[Collection[ColumnHeader]]=None) -> List[str]:
        if column_headers is None:
            return list(self.column_id_to_column_header[sheet_index].keys())
        try:
            return [
                self.column_header_to_column_id[sheet_index][column_header] 
                for column_header in column_headers
            ]
        except:
            raise make_no_column_error(column_headers)

    def get_column_headers(self, sheet_index: int) -> List[ColumnHeader]:
        return list(self.column_id_to_column_header[sheet_index].values())

    def get_column_ids_map(self, sheet_index: int) -> Dict[str, ColumnHeader]:
        return self.column_id_to_column_header[sheet_index]

    def get_column_id_by_header(self, sheet_index: int, column_header: ColumnHeader) -> str:
        # We handle nan values here because nan values are not hashable, and so
        # we can't use them as keys in the column_header_to_column_id dict
        if isinstance(column_header, float) and np.isnan(column_header):
            for header, column_id in self.column_header_to_column_id[sheet_index].items():
                if isinstance(header, float) and np.isnan(header):
                    return column_id

        return self.column_header_to_column_id[sheet_index][column_header]

    def get_column_ids_by_headers(self, sheet_index: int, column_headers: List[ColumnHeader]) -> List[ColumnID]:
        return [
            self.column_header_to_column_id[sheet_index][column_header] 
            for column_header in column_headers
        ]

    def get_column_header_by_id(self, sheet_index: int, column_id: ColumnID) -> ColumnHeader:
        return self.column_id_to_column_header[sheet_index][column_id]

    def get_column_headers_by_ids(self, sheet_index: int, column_ids: List[ColumnID]) -> List[ColumnHeader]:
        return [
            self.column_id_to_column_header[sheet_index][column_id] 
            for column_id in column_ids
        ]

    def move_to_deprecated_id_format(self) -> None:
        """
        Helper function to change the column IDs, specifically for the 
        purpose of backwards compability. This simulates a preprocessing 
        step.
        """
        from mitosheet.step_performers.bulk_old_rename.deprecated_utils import \
            make_valid_header
        for sheet_index in range(len(self.column_id_to_column_header)):
            column_id_to_column_header_map = self.column_id_to_column_header[sheet_index]

            new_column_id_to_column_header_map = dict()
            new_column_header_to_column_id_map = dict()

            for column_id, column_header in column_id_to_column_header_map.items():
                deprecated_column_id = make_valid_header(column_id)
                new_column_id_to_column_header_map[deprecated_column_id] = column_header
                new_column_header_to_column_id_map[column_header] = deprecated_column_id
            
            self.column_id_to_column_header[sheet_index] = new_column_id_to_column_header_map
            self.column_header_to_column_id[sheet_index] = new_column_header_to_column_id_map

    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
When we switched from column headers to column ids in our 
analyses, we needed to upgrade a bunch of different steps.

We have common utilities here for supporting in this task.
"""
from typing import Any, Dict
from mitosheet.column_headers import get_column_header_id, get_column_header_ids

def replace_headers_with_id(params: Dict[str, Any], old_key: str, new_key: str) -> Dict[str, Any]:
    """
    Helper function for upgrading from column headers 
    to column ids in the params of a step. Handles the
    various data types that the steps are in.
    """

    old_value_column_header = params[old_key]
    new_value: Any = old_value_column_header
    if isinstance(old_value_column_header, str):
        new_value = get_column_header_id(old_value_column_header)
    elif isinstance(old_value_column_header, list):
        new_value = get_column_header_ids(old_value_column_header)
    elif isinstance(old_value_column_header, dict):
        new_value = {
            get_column_header_id(key): value
            for key, value in old_value_column_header.items()
        }
    
    del params[old_key]
    params[new_key] = new_value

    return params



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.types import ColumnID, UserDefinedImporterParamType
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_to_transpiled_code

def get_transpiled_importer_params(user_defined_importer_params: Dict[str, Any]) -> str:
    param_strings = []
    for param_name, param_value in user_defined_importer_params.items():
        param_strings.append(f'{param_name}={column_header_to_transpiled_code(param_value)}')
    return ", ".join(param_strings)



class UserDefinedImportCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, importer: str, user_defined_importer_params: Dict[str, Any]):
        super().__init__(prev_state, post_state)
        self.importer = importer
        self.user_defined_importer_params = user_defined_importer_params

        self.df_names = [df_name for index, df_name in enumerate(self.post_state.df_names) if index >= len(self.prev_state.df_names)]

    def get_display_name(self) -> str:
        return 'User Defined Import'
    
    def get_description_comment(self) -> str:
        return f"Imported {', '.join(self.df_names)} using {self.importer}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        # For each new dataframe, we get it's name
        new_df_names = self.post_state.df_names[len(self.prev_state.df_names):]
        df_name_string = ', '.join(new_df_names)
        code = f"{df_name_string} = {self.importer}({get_transpiled_importer_params(self.user_defined_importer_params)})"
        return [code], []
    
    def get_created_sheet_indexes(self) -> Optional[List[int]]:
        num_new_dfs = len(self.post_state.dfs) - len(self.prev_state.dfs)
        return [i for i in range(len(self.post_state.dfs) - num_new_dfs, len(self.post_state.dfs))]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.no_op_code_chunk import NoOpCodeChunk
from mitosheet.code_chunks.step_performers.column_steps.rename_columns_code_chunk import RenameColumnsCodeChunk

from mitosheet.errors import make_column_exists_error, raise_error_if_column_ids_do_not_exist
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnHeader, ColumnID


class RenameColumnStepPerformer(StepPerformer):
    """"
    A rename_column step, which allows you to rename a column
    in a dataframe.

    NOTE: this should only be called on dataframes that do
    not have multi-index headers!
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'rename_column' 

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')
        new_column_header: str = get_param(params, 'new_column_header')
        level: int = get_param(params, 'new_column_header')

        raise_error_if_column_ids_do_not_exist(
            'rename column',
            prev_state,
            sheet_index,
            column_id
        )

        if new_column_header in prev_state.dfs[sheet_index].keys():
            raise make_column_exists_error(new_column_header)

        # If the user has deleted the column header entirely, this is very likely
        # a mistake and not something they meant to do... so we just don't make any edits
        # and don't throw an error
        if new_column_header == '':
            return prev_state, None

        # Create a new post state for this step
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        old_level_value, pandas_processing_time = rename_column_headers_in_state(
            post_state,
            sheet_index,
            column_id,
            new_column_header,
            level
        )

        return post_state, {
            'old_level_value': old_level_value,
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        if params['new_column_header'] == '':
            # If the new column header is an empty string, it's a noop
            return [NoOpCodeChunk(prev_state, post_state)]


        return [
            RenameColumnsCodeChunk(
                prev_state, 
                post_state, 
                # We construct a rename for mulitple columns, as this is the most 
                # convenient way to allow us to combine multiple renames
                # into one
                params['sheet_index'],
                {
                    params['column_id']: params['new_column_header']
                },
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


def rename_column_headers_in_state(
        post_state: State,
        sheet_index: int,
        column_id: ColumnID,
        new_column_header: ColumnHeader,
        level: Optional[int]=None
    ) -> Tuple[Optional[ColumnHeader], float]:
    """
    A helper function for updating a column header in the state, which is useful
    for both this rename step and for the bulk rename step.
    """
    old_column_header = post_state.column_ids.get_column_header_by_id(sheet_index, column_id)

    # If the level is not set, just do a simple rename
    post_state.dfs[sheet_index].rename(columns={old_column_header: new_column_header}, inplace=True)
    
    # Update the column header
    pandas_start_time = perf_counter()
    post_state.column_ids.set_column_header(sheet_index, column_id, new_column_header)
    pandas_processing_time = perf_counter() - pandas_start_time
    
    return old_column_header, pandas_processing_time


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State

from mitosheet.transpiler.transpile_utils import (
    column_header_list_to_transpiled_code, column_header_to_transpiled_code)
from mitosheet.types import ColumnID

class DropDuplicatesCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_ids: List[ColumnID], keep: str):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.column_ids = column_ids
        self.keep = keep

        self.df_name = self.post_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Dropped duplicates'
    
    def get_description_comment(self) -> str:
        return f'Dropped duplicates in {self.df_name}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        column_headers = [
            self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, column_id)
            for column_id in self.column_ids
        ]

        # If the subset is none, then we don't actually do the drop, as there are no
        # duplicates between 0 columns
        if len(column_headers) == 0:
            return [], []
        
        # We leave subset and keep empty if they are not used
        param_string = ''
        if len(column_headers) != len(self.post_state.dfs[self.sheet_index].keys()):
            param_string += 'subset=' + column_header_list_to_transpiled_code(column_headers) + ', '
        
        param_string += 'keep=' + column_header_to_transpiled_code(self.keep) # not a column header, but we can use the same utility

        return [
            f'{self.df_name} = {self.df_name}.drop_duplicates({param_string})'
        ], []
    
    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from typing import Any, Dict, List, Optional, Tuple, Union

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.excel_utils import (get_column_from_column_index,
                                   get_col_and_row_indexes_from_range)
from mitosheet.public.v2.excel_utils import get_read_excel_params_from_range
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_to_transpiled_code, param_dict_to_code
from mitosheet.types import ExcelRangeImport



EXCEL_RANGE_IMPORT_TYPE_RANGE = 'range'

EXCEL_SHEET_TYPE_SHEET_NAME = 'sheet name'
EXCEL_SHEET_TYPE_SHEET_INDEX = 'sheet index'

EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE = 'upper left corner value'
EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE_STARTS_WITH = 'upper left corner value starts with'
EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE_CONTAINS = 'upper left corner value contains'
EXCEL_RANGE_START_CONDITIONS = [
    EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE,
    EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE_STARTS_WITH,
    EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE_CONTAINS
]

EXCEL_RANGE_END_CONDITION_FIRST_EMPTY_VALUE = 'first empty cell'
EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE = 'bottom left corner value'
EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE_STARTS_WITH = 'bottom left corner value starts with'
EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE_CONTAINS = 'bottom left corner value contains'
EXCEL_RANGE_END_CONDITIONAL_BOTTOM_LEFT_CORNER_CONSECUTIVE_EMPTY_CELLS = 'bottom left corner consecutive empty cells'
EXCEL_RANGE_END_CONDITIONAL_BOTTOM_LEFT_CORNER_CONSECUTIVE_EMPTY_CELLS_IN_FIRST_COLUMN = 'bottom left corner consecutive empty cells in first column'
EXCEL_RANGE_END_CONDITION_ROW_ENTIRELY_EMPTY = 'row entirely empty'
EXCEL_RANGE_END_CONDITION_CUMULATIVE_NUMBER_OF_EMPTY_ROWS = 'cumulative number of empty rows'
EXCEL_RANGE_END_CONDTIONS = [
    EXCEL_RANGE_END_CONDITION_FIRST_EMPTY_VALUE,
    EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE,
    EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE_STARTS_WITH,
    EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE_CONTAINS,
    EXCEL_RANGE_END_CONDITIONAL_BOTTOM_LEFT_CORNER_CONSECUTIVE_EMPTY_CELLS,
    EXCEL_RANGE_END_CONDITIONAL_BOTTOM_LEFT_CORNER_CONSECUTIVE_EMPTY_CELLS_IN_FIRST_COLUMN,
    EXCEL_RANGE_END_CONDITION_ROW_ENTIRELY_EMPTY,
    EXCEL_RANGE_END_CONDITION_CUMULATIVE_NUMBER_OF_EMPTY_ROWS
]

EXCEL_RANGE_COLUMN_END_CONDITION_FIRST_EMPTY_CELL = 'first empty cell'
EXCEL_RANGE_COLUMN_END_CONDITION_NUM_COLUMNS = 'num columns'
EXCEL_RANGE_COLUMN_END_CONDITIONS = [
    EXCEL_RANGE_COLUMN_END_CONDITION_FIRST_EMPTY_CELL,
    EXCEL_RANGE_COLUMN_END_CONDITION_NUM_COLUMNS,
]

def get_table_range_params(sheet: Dict[str, Union[str, int]], start_condition: Any, end_condition: Any, column_end_condition: Any) -> Dict[str, Any]:
    """
    Get the params for a get_table_range call, not including the file_path param, as we transpile that specially,
    because it might have to be a variable.
    """

    if start_condition['type'] not in EXCEL_RANGE_START_CONDITIONS:
        raise ValueError(f'Invalid start condition type: {start_condition["type"]}')
    if end_condition['type'] not in EXCEL_RANGE_END_CONDTIONS:
        raise ValueError(f'Invalid end condition type: {end_condition["type"]}')
    if column_end_condition['type'] not in EXCEL_RANGE_COLUMN_END_CONDITIONS:
        raise ValueError(f'Invalid column end condition type: {column_end_condition["type"]}')
    
    sheet_type = sheet['type']
    if sheet_type == EXCEL_SHEET_TYPE_SHEET_NAME:
        sheet_name = sheet['value']
        sheet_index = None
    elif sheet_type == EXCEL_SHEET_TYPE_SHEET_INDEX:
        sheet_name = None
        sheet_index = sheet['value']
    else:
        raise ValueError(f'Invalid sheet type: {sheet_type}')

    upper_left_value = start_condition['value'] if start_condition['type'] == EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE else None
    upper_left_value_starts_with = start_condition['value'] if start_condition['type'] == EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE_STARTS_WITH else None
    upper_left_value_contains = start_condition['value'] if start_condition['type'] == EXCEL_RANGE_START_CONDITION_UPPER_LEFT_VALUE_CONTAINS else None

    bottom_left_value = end_condition['value'] if end_condition['type'] == EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE else None
    bottom_left_value_starts_with = end_condition['value'] if end_condition['type'] == EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE_STARTS_WITH else None
    bottom_left_value_contains = end_condition['value'] if end_condition['type'] == EXCEL_RANGE_END_CONDITION_BOTTOM_LEFT_CORNER_VALUE_CONTAINS else None
    bottom_left_consecutive_empty_cells = end_condition['value'] if end_condition['type'] == EXCEL_RANGE_END_CONDITIONAL_BOTTOM_LEFT_CORNER_CONSECUTIVE_EMPTY_CELLS else None
    bottom_left_consecutive_empty_cells_in_first_column = end_condition['value'] if end_condition['type'] == EXCEL_RANGE_END_CONDITIONAL_BOTTOM_LEFT_CORNER_CONSECUTIVE_EMPTY_CELLS_IN_FIRST_COLUMN else None
    row_entirely_empty = True if end_condition['type'] == EXCEL_RANGE_END_CONDITION_ROW_ENTIRELY_EMPTY else None
    cumulative_number_of_empty_rows = end_condition['value'] if end_condition['type'] == EXCEL_RANGE_END_CONDITION_CUMULATIVE_NUMBER_OF_EMPTY_ROWS else None

    num_columns = column_end_condition['value'] if column_end_condition['type'] == EXCEL_RANGE_COLUMN_END_CONDITION_NUM_COLUMNS else None

    all_params = {
        'sheet_name': sheet_name,
        'sheet_index': sheet_index,
        'upper_left_value': upper_left_value,
        'upper_left_value_starts_with': upper_left_value_starts_with,
        'upper_left_value_contains': upper_left_value_contains,
        'bottom_left_corner_consecutive_empty_cells': bottom_left_consecutive_empty_cells,
        'bottom_left_consecutive_empty_cells_in_first_column': bottom_left_consecutive_empty_cells_in_first_column,
        'bottom_left_value': bottom_left_value,
        'bottom_left_value_starts_with': bottom_left_value_starts_with,
        'bottom_left_value_contains': bottom_left_value_contains,
        'row_entirely_empty': row_entirely_empty,
        'cumulative_number_of_empty_rows': cumulative_number_of_empty_rows,
        'num_columns': num_columns
    }

    # Return only non-None params
    return {k: v for k, v in all_params.items() if v is not None}


class ExcelRangeImportCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, file_path: str, sheet: Dict[str, Union[int, str]], range_imports: List[ExcelRangeImport], convert_csv_to_xlsx: bool):
        super().__init__(prev_state, post_state)
        self.file_path = file_path
        self.sheet = sheet
        self.range_imports = range_imports
        self.convert_csv_to_xlsx = convert_csv_to_xlsx

    def get_display_name(self) -> str:
        return 'Excel Range Import'
    
    def get_description_comment(self) -> str:

        if self.sheet['type'] == EXCEL_SHEET_TYPE_SHEET_NAME:
            sheet = f"{self.sheet['value']}"
        else:
            sheet = f"sheet at index {self.sheet['value']}"
        
        return f"Imported {len(self.range_imports)} dataframes from {sheet} in {self.file_path}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        code = []

        transpiled_sheet_name = column_header_to_transpiled_code(self.sheet['value'])

        if self.convert_csv_to_xlsx:
            code.append(f'xlsx_file_path = convert_csv_file_to_xlsx_file(\'{self.file_path}\', {transpiled_sheet_name})')
            transpiled_file_path = f'xlsx_file_path'
        else:
            transpiled_file_path = f'r\'{self.file_path}\''


        for idx, range_import in enumerate(self.range_imports):

            sheet_index = len(self.prev_state.dfs) + idx
            df_name = self.post_state.df_names[sheet_index]

            # If it's an explicit range, then just import that exact range
            if range_import['type'] == EXCEL_RANGE_IMPORT_TYPE_RANGE:
                _range = range_import['value'] #type: ignore
                skiprows, nrows, usecols = get_read_excel_params_from_range(_range)
                
                code.append(
                    f'{df_name} = pd.read_excel({transpiled_file_path}, sheet_name={transpiled_sheet_name}, skiprows={skiprows}, nrows={nrows}, usecols=\'{usecols}\')'
                )

            else:
                # Otherwise, if you're importing based on values, we generate dynamic code
                start_condition = range_import['start_condition'] # type: ignore
                end_condition = range_import['end_condition'] #type: ignore
                column_end_condition = range_import['column_end_condition'] #type: ignore

                params = get_table_range_params(self.sheet, start_condition, end_condition, column_end_condition)
                params_code = param_dict_to_code(params, as_single_line=True)

                code.extend([
                    f'_range = get_table_range({transpiled_file_path}, {params_code})',
                    'skiprows, nrows, usecols = get_read_excel_params_from_range(_range)',
                    f'{df_name} = pd.read_excel({transpiled_file_path}, sheet_name={transpiled_sheet_name}, skiprows=skiprows, nrows=nrows, usecols=usecols)'
                ])
                

            # Add a new line in between different imports otherwise the code looks bad
            if idx < len(self.range_imports) - 1:
                code.append('')

        return code, ['import pandas as pd']

    def get_created_sheet_indexes(self) -> Optional[List[int]]:
        return [i for i in range(len(self.post_state.dfs) - len(self.range_imports), len(self.post_state.dfs))]

    def _combine_right_with_excel_range_import_code_chunk(self, excel_range_import_code_chunk: "ExcelRangeImportCodeChunk") -> Optional[CodeChunk]:
        if excel_range_import_code_chunk.file_path == self.file_path and excel_range_import_code_chunk.sheet == self.sheet:
            new_range_imports = copy(self.range_imports)
            new_range_imports.extend(excel_range_import_code_chunk.range_imports)

            return ExcelRangeImportCodeChunk(
                self.prev_state,
                excel_range_import_code_chunk.post_state,
                self.file_path,
                self.sheet,
                new_range_imports,
                self.convert_csv_to_xlsx
            )

        return None


    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, ExcelRangeImportCodeChunk):
            return self._combine_right_with_excel_range_import_code_chunk(other_code_chunk)

        return None
    
    def get_parameterizable_params(self) -> List[Tuple[str, str, str]]:
        return [(f'r{column_header_to_transpiled_code(self.file_path)}', 'file_name', 'Excel import file path')]
    



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_list_to_transpiled_code

class DeleteRowCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, labels: List[Any]):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.labels = labels

        self.df_name = self.post_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Delete Row'
    
    def get_description_comment(self) -> str:
        return f'Deleted {len(self.labels)} row{"" if len(self.labels) == 1 else "s"} in {self.df_name}'
        
    def get_code(self) -> Tuple[List[str], List[str]]:
        return [f'{self.df_name}.drop(labels={column_header_list_to_transpiled_code(self.labels)}, inplace=True)'], []
    
    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def _combine_right_with_delete_row_code_chunk(self, other_code_chunk: "DeleteRowCodeChunk") -> Optional["DeleteRowCodeChunk"]:
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None
        
        all_labels = copy(self.labels) # Make sure to copy this so we don't get weird bugs w/ duplication
        all_labels.extend(other_code_chunk.labels)
        
        return DeleteRowCodeChunk(
            self.prev_state,
            other_code_chunk.post_state,
            self.sheet_index,
            all_labels
        )

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, DeleteRowCodeChunk):
            return self._combine_right_with_delete_row_code_chunk(other_code_chunk)
            
        return None
    


#!/usr/bin/env python
# coding: utf-8
#
# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import os
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.import_steps.excel_import_code_chunk import \
    ExcelImportCodeChunk, build_read_excel_params
from mitosheet.errors import make_file_not_found_error
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.utils import get_valid_dataframe_name


class ExcelImportStepPerformer(StepPerformer):
    """
    A simple import, which allows you to import excel files 
    with the given file_name.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'excel_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        file_name: str = get_param(params, 'file_name')
        sheet_names: List[str] = get_param(params, 'sheet_names')
        has_headers: bool = get_param(params, 'has_headers')
        skiprows: int = get_param(params, 'skiprows')
        decimal: str = get_param(params, 'decimal')

        read_excel_params = build_read_excel_params(sheet_names, has_headers, skiprows, decimal)

        post_state = prev_state.copy()
        
        if not os.path.exists(file_name):
            raise make_file_not_found_error(file_name)

        pandas_start_time = perf_counter()
        df_dictonary = pd.read_excel(file_name, **read_excel_params, engine='openpyxl') 
        pandas_processing_time = perf_counter() - pandas_start_time

        for sheet_name, df in df_dictonary.items():
            post_state.add_df_to_state(
                df, 
                DATAFRAME_SOURCE_IMPORTED, 
                df_name=get_valid_dataframe_name(post_state.df_names, sheet_name),
            )

        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ExcelImportCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'file_name'),
                get_param(params, 'sheet_names'),
                get_param(params, 'has_headers'),
                get_param(params, 'skiprows'),
                get_param(params, 'decimal')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
For going to a datetime series.
"""


from typing import Any
from mitosheet.public.v1.sheet_functions.types.utils import get_datetime_format, get_to_datetime_params

import numpy as np
import pandas as pd

from mitosheet.is_type_utils import (is_bool_dtype,
                                                   is_datetime_dtype,
                                                   is_number_dtype,
                                                   is_string_dtype)


def to_datetime_series(
        unknown_object: Any,
        on_uncastable_arg_element: Any=('default', np.NaN), # Union[Literal['error'], Tuple[Literal['default'], any]]
    ) -> pd.Series:
    """
    Converts the given object to a datetime series.
    """

    # If it is not a series, we put it in a series, and get the type again
    if not isinstance(unknown_object, pd.Series):
        unknown_object = pd.Series([unknown_object])

    column_dtype = str(unknown_object.dtype)
    if is_bool_dtype(column_dtype):
        return None
    elif is_datetime_dtype(column_dtype):
        return unknown_object
    if is_number_dtype(column_dtype):
        return None
    elif is_string_dtype(column_dtype):
        # TODO: improve this to work element wise!
        to_datetime_params = get_to_datetime_params(unknown_object)
        return pd.to_datetime(
            unknown_object,
            errors='coerce',
            **to_datetime_params
        )
    
    else:
        return None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
from typing import Any, Dict, List
import pandas as pd
from mitosheet.types import StepsManagerType
from mitosheet.user.location import is_streamlit


NO_DEFINED_DF_MESSAGE = 'No variables match your requested type.'

def get_df_names_ipython() -> List[str]:
    from IPython import get_ipython
    from io import StringIO 
    import sys

    ipython = get_ipython() # type: ignore

    class Capturing(list):
        def __enter__(self):
            self._stdout = sys.stdout
            sys.stdout = self._stringio = StringIO()
            return self
        def __exit__(self, *args):
            self.extend(self._stringio.getvalue().splitlines())
            del self._stringio    # free up some memory
            sys.stdout = self._stdout

    with Capturing() as output:
        ipython.run_line_magic("who",  "DataFrame")
        
    output = [df.strip() for line in output for df in line.split("\t") if df.strip() != '']
    
    # If we get a message there are no variables, clear this out of the ouput
    if len(output) == 1 and output[0] == NO_DEFINED_DF_MESSAGE:
        output = []

    return output

def get_df_names_streamlit() -> List[str]:
    # Get dataframes defined in any calling frames above this one
    import inspect
    import streamlit as st
    import pandas as pd

    # Get the calling frame
    dfs = []
    frame = inspect.currentframe()
    while frame is not None:
        if frame.f_code.co_name == 'write':
            break
        frame = frame.f_back

        # Get the calling frame's locals
        if frame is None:
            break
        frame_locals = frame.f_locals
        # Get the dataframes defined in the calling frame
        for name, value in frame_locals.items():
            if isinstance(value, pd.DataFrame):
                dfs.append(name)

    return dfs



def get_defined_df_names(params: Dict[str, Any], steps_manager: StepsManagerType) -> List[str]:
    if is_streamlit():
        return get_df_names_streamlit()
    return get_df_names_ipython()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict
from mitosheet.types import StepsManagerType


def get_dataframe_as_csv(params: Dict[str, Any], steps_manager: StepsManagerType) -> str:
    """
    Sends a dataframe as a CSV string
    """
    sheet_index = params['sheet_index']
    df = steps_manager.dfs[sheet_index]

    return df.to_csv(index=False)





from datetime import datetime, timedelta
from typing import Optional, Union

import numpy as np

def cast_to_string(unknown: Union[str, int, float, bool, datetime, timedelta]) -> Optional[str]:
    if isinstance(unknown, float) and np.isnan(unknown):
        return None

    return str(unknown)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import pandas as pd
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.ai.ai_utils import fix_final_dataframe_name, fix_up_missing_imports
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.ai_transformation_code_chunk import AITransformationCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.ai.recon import exec_and_get_new_state_and_result

class AITransformationStepPerformer(StepPerformer):
    """
    Allows you to execute an arbitrary chunk of code, generated by an AI.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'ai_transformation'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        # We don't use any of these parameters, but we keep them for clarity, and we don't want to remove them accidently
        user_input: str = get_param(params, 'user_input')
        prompt_version: str = get_param(params, 'prompt_version')
        prompt: str = get_param(params, 'prompt')
        completion: str = get_param(params, 'completion')
        edited_completion: str = get_param(params, 'edited_completion')

        pandas_start_time = perf_counter()
        post_state, last_line_value, frontend_result = exec_and_get_new_state_and_result(prev_state, edited_completion)
        pandas_processing_time = perf_counter() - pandas_start_time

        if isinstance(last_line_value, pd.DataFrame) or isinstance(last_line_value, pd.Series):
            final_code = fix_final_dataframe_name(edited_completion, post_state.df_names[-1], isinstance(last_line_value, pd.Series))
        else:
            final_code = edited_completion

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'result': frontend_result,
            'final_code': final_code
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            AITransformationCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'user_input'),
                get_param(execution_data if execution_data is not None else {}, 'final_code')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return set() # NOTE: We act as though we modify all sheet indexes, as we can't easily figure out what we do modify

    @classmethod
    def get_created_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1} # NOTE: We act as though we create all sheet indexes, as we can't easily figure out what we do modify


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import os
from mitosheet.transpiler.transpile_utils import NEWLINE_TAB, TAB
import pytest
import pandas as pd

from mitosheet.api.get_parameterizable_params import get_parameterizable_params
from mitosheet.transpiler.transpile import transpile
from mitosheet.tests.test_utils import create_mito_wrapper_with_data, create_mito_wrapper
from mitosheet.tests.decorators import pandas_post_1_2_only, python_post_3_6_only

def test_transpile_single_column():
    mito = create_mito_wrapper_with_data(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        "df1.insert(1, 'B', df1[\'A\'])", 
        ''
    ]


def test_transpile_multiple_columns_no_relationship():
    mito = create_mito_wrapper_with_data(['abc'])
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')
    print(mito.transpiled_code)
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1.insert(1, \'B\', 0)', 
        '',
        'df1.insert(2, \'C\', 0)', 
        ''
    ]

def test_transpile_columns_in_each_sheet():
    mito = create_mito_wrapper_with_data(['abc'], sheet_two_A_data=['abc'])
    mito.add_column(0, 'B')
    mito.add_column(1, 'B')

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1.insert(1, \'B\', 0)',
        '',
        'df2.insert(1, \'B\', 0)',
        ''
    ]

def test_transpile_multiple_columns_linear():
    mito = create_mito_wrapper_with_data(['abc'])
    mito.set_formula('=A', 0, 'B', add_column=True)
    mito.set_formula('=B', 0, 'C', add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1.insert(1, \'B\', df1[\'A\'])',
        '',
        'df1.insert(2, \'C\', df1[\'B\'])',
        '',
    ]

COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet(column_header):
    mito = create_mito_wrapper_with_data(['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        f'df1.insert(1, \'{column_header}\', df1[\'A\'])', 
        '',
    ]


COLUMN_HEADERS = [
    ('ABC'),
    ('ABC_D'),
    ('ABC_DEF'),
    ('ABC_123'),
    ('ABC_HAHA_123'),
    ('ABC_HAHA-123'),
    ('---data---'),
    ('---da____ta---'),
    ('--'),
]
@pytest.mark.parametrize("column_header", COLUMN_HEADERS)
def test_transpile_column_headers_non_alphabet_multi_sheet(column_header):
    mito = create_mito_wrapper_with_data(['abc'], sheet_two_A_data=['abc'])
    mito.set_formula('=A', 0, column_header, add_column=True)
    mito.set_formula('=A', 1, column_header, add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        f'df1.insert(1, \'{column_header}\', df1[\'A\'])', 
        '',
        f'df2.insert(1, \'{column_header}\', df2[\'A\'])', 
        '',
    ]

def test_preserves_order_columns():
    mito = create_mito_wrapper_with_data(['abc'])
    # Topological sort will currently display this in C, B order
    mito.add_column(0, 'B')
    mito.add_column(0, 'C')

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1.insert(1, \'B\', 0)',
        '',
        'df1.insert(2, \'C\', 0)',
        '',
    ]

def test_transpile_delete_columns():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper(df1)
    mito.delete_columns(0, ['C', 'B'])

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1.drop([\'C\', \'B\'], axis=1, inplace=True)',
        '',
    ]


# TESTING OPTIMIZATION

def test_removes_unedited_formulas_for_unedited_sheets():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper(df1, df2)

    mito.set_formula('=C', 0, 'D', add_column=True)
    mito.set_formula('=C', 1, 'D', add_column=True)

    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C', 'D'], ['A', 'B', 'C', 'D'])

    mito.set_formula('=C + 1', 1, 'D', add_column=True)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        "df1.insert(3, 'D', df1[\'C\'])", 
        '',
        "df2.insert(3, 'D', df2[\'C\'])", 
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
        'df2[\'D\'] = df2[\'C\'] + 1',
        '',
    ]


def test_mulitple_merges_no_formula_steps():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper(df1, df2)
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C'], ['A', 'B', 'C'])


    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df4 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df5 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
    ]

def test_optimization_with_other_edits():
    df1 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    df2 = pd.DataFrame(data={'A': [1], 'B': [101], 'C': [11]})
    mito = create_mito_wrapper(df1, df2)
    mito.add_column(0, 'D')
    mito.set_formula('=A', 0, 'D')
    mito.merge_sheets('lookup', 0, 1, [['A', 'A']], ['A', 'B', 'C', 'D'], ['A', 'B', 'C'])
    mito.add_column(0, 'AAA')
    mito.delete_columns(0, ['AAA'])

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        "df1.insert(3, 'D', df1[\'A\'])", 
        '',
        'temp_df = df2.drop_duplicates(subset=[\'A\']) # Remove duplicates so lookup merge only returns first match', 
        'df3 = df1.merge(temp_df, left_on=[\'A\'], right_on=[\'A\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
    ]


def test_transpile_does_no_initial():
    df1 = pd.DataFrame(data={'First Name': ['Nate', 'Nate'], 123: ['Rush', 'Jack'], True: ['1', '2']})
    mito = create_mito_wrapper(df1)

    assert len(mito.transpiled_code) == 0

    
def test_transpile_reorder_column():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper(df1)
    mito.reorder_column(0, 'A', 1)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        '',
    ]

def test_transpile_two_column_reorders():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper(df1)
    mito.reorder_column(0, 'A', 1)
    mito.reorder_column(0, 'B', 1)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        '',
        'df1_columns = [col for col in df1.columns if col != \'B\']',
        'df1_columns.insert(1, \'B\')',
        'df1 = df1[df1_columns]',
        '',
    ]

def test_transpile_reorder_column_invalid():
    df1 = pd.DataFrame(data={'A': ['aaron'], 'B': ['jon']})
    mito = create_mito_wrapper(df1)
    mito.reorder_column(0, 'A', 5)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'df1_columns = [col for col in df1.columns if col != \'A\']',
        'df1_columns.insert(1, \'A\')',
        'df1 = df1[df1_columns]',
        '',
    ]

def test_transpile_merge_then_sort():
    df1 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Number': [123, 1]})
    df2 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Sign': ['Gemini', "Tarus"]})
    mito = create_mito_wrapper(df1, df2)
    mito.merge_sheets('lookup', 0, 1, [['Name', 'Name']], list(df1.keys()), list(df2.keys()))
    mito.sort(2, 'Number', 'ascending')

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'temp_df = df2.drop_duplicates(subset=[\'Name\']) # Remove duplicates so lookup merge only returns first match',
        'df3 = df1.merge(temp_df, left_on=[\'Name\'], right_on=[\'Name\'], how=\'left\', suffixes=[\'_df1\', \'_df2\'])',
        '',
        'df3 = df3.sort_values(by=\'Number\', ascending=True, na_position=\'first\')',
        '',
    ]

def test_transpile_multiple_pandas_imports_combined(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame(data={'Name': ["Aaron", "Nate"], 'Number': [123, 1]})
    df1.to_csv(tmp_file, index=False)
    mito = create_mito_wrapper(df1)
    mito.simple_import([tmp_file])
    mito.add_column(0, 'A', -1)
    mito.simple_import([tmp_file])
    mito.add_column(1, 'A', -1)
    mito.simple_import([tmp_file])

    assert len(mito.optimized_code_chunks) == 5
    assert 'import pandas as pd' in mito.transpiled_code
    assert len([c for c in mito.transpiled_code if c == 'import pandas as pd']) == 1

def test_transpile_as_function_no_params(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    print(mito.transpiled_code)

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function():",
        f"{TAB}txt = pd.read_csv(r'{tmp_file}')",
        f'{TAB}',
        f"{TAB}return txt",
        "",
        "txt = function()"
    ]

def test_transpile_as_function_no_call(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.code_options_update_no_check_transpiled({'as_function': True, 'call_function': False, 'function_name': 'function', 'function_params': {}})


    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function():",
        f"{TAB}txt = pd.read_csv(r'{tmp_file}')",
        f'{TAB}',
        f"{TAB}return txt",
        ""
    ]

def test_transpile_as_function_df_params():
    mito = create_mito_wrapper(pd.DataFrame({'A': [1]}), arg_names=['df1'])
    mito.add_column(0, 'B')
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    print(mito.transpiled_code)
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        '',
        'def function(df1):',
        f"{TAB}df1.insert(1, 'B', 0)",
        f'{TAB}',
        f"{TAB}return df1",
        "",
        "df1 = function(df1)"
    ]

def test_transpile_as_function_string_params():
    tmp_file = 'txt.csv'
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper(str(tmp_file), arg_names=[f"'{tmp_file}'"])
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        'def function(txt_path):',
        f"{TAB}# Read in filepaths as dataframes",
        f"{TAB}txt = pd.read_csv(txt_path)",
        f'{TAB}',
        f"{TAB}return txt",
        "",
        "txt = function('txt.csv')"
    ]

    os.remove(tmp_file)

def test_transpile_as_function_string_params_no_args_update():
    tmp_file = 'txt.csv'
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper(str(tmp_file))
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        'def function(txt_path):',
        f"{TAB}# Read in filepaths as dataframes",
        f"{TAB}txt = pd.read_csv(txt_path)",
        f'{TAB}',
        f"{TAB}return txt",
        "",
        'txt = function("txt.csv")'
    ]

    os.remove(tmp_file)

def test_transpile_as_function_both_params():
    tmp_file = 'txt.csv'
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper(df1, str(tmp_file), arg_names=['df1', f"'{tmp_file}'"])
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        'def function(df1, txt_path):',
        f"{TAB}# Read in filepaths as dataframes",
        f"{TAB}txt = pd.read_csv(txt_path)",
        f'{TAB}',
        f"{TAB}return df1, txt",
        "",
        "df1, txt = function(df1, 'txt.csv')"
    ]

    os.remove(tmp_file)


def test_transpile_pivot_table_indents():
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Nate'], 'Height': [4, 5]})
    mito = create_mito_wrapper(df1, arg_names=['df1'])

    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    mito.pivot_sheet(
        0, 
        ['Name'],
        [],
        {'Height': ['sum']}
    )
    
    print("\n".join(mito.transpiled_code))
    assert "    pivot_table = tmp_df.pivot_table(\n        index=['Name'],\n        values=['Height'],\n        aggfunc={'Height': ['sum']}\n    )" in mito.transpiled_code


def test_transpile_as_function_single_param(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name):",
        f"{TAB}txt = pd.read_csv(var_name)",
        f'{TAB}',
        f"{TAB}return txt",
        "",
        f"var_name = r'{tmp_file}'",
        "",
        f"txt = function(var_name)"
    ]


def test_transpile_as_function_both_params_and_additional():
    tmp_file = 'txt.csv'
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper(df1, str(tmp_file), arg_names=['df1', f"'{tmp_file}'"])
    mito.simple_import([tmp_file])
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        'import pandas as pd',
        '',
        'def function(df1, txt_path, var_name):',
        f"{TAB}# Read in filepaths as dataframes",
        f"{TAB}txt = pd.read_csv(txt_path)",
        f'{TAB}',
        '    txt_1 = pd.read_csv(var_name)',
        f'{TAB}',
        f"{TAB}return df1, txt, txt_1",
        "",
        f"var_name = r'{tmp_file}'",
        "",
        f"df1, txt, txt_1 = function(df1, 'txt.csv', var_name)"
    ]

    os.remove(tmp_file)

def test_transpile_as_function_single_param_multiple_times(tmp_path):
    tmp_file = str(tmp_path / 'txt.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file])
    mito.simple_import([tmp_file])
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name):",
        f"{TAB}txt = pd.read_csv(var_name)",
        f"{TAB}txt_1 = pd.read_csv(var_name)",
        f'{TAB}',
        f"{TAB}return txt, txt_1",
        "",
        f"var_name = r'{tmp_file}'",
        "",
        f"txt, txt_1 = function(var_name)"
    ]

def test_transpile_as_function_multiple_params(tmp_path):
    tmp_file1 = str(tmp_path / 'txt.csv')
    tmp_file2 = str(tmp_path / 'file.csv')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_csv(tmp_file1, index=False)
    df1.to_csv(tmp_file2, index=False)

    mito = create_mito_wrapper()
    mito.simple_import([tmp_file1])
    mito.simple_import([tmp_file2])
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name1': f"r'{tmp_file1}'", 'var_name2': f"r'{tmp_file2}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name1, var_name2):",
        f"{TAB}txt = pd.read_csv(var_name1)",
        f"{TAB}file = pd.read_csv(var_name2)",
        f'{TAB}',
        f"{TAB}return txt, file",
        "",
        f"var_name1 = r'{tmp_file1}'",
        f"var_name2 = r'{tmp_file2}'",
        "",
        f"txt, file = function(var_name1, var_name2)"
    ]

@pandas_post_1_2_only
@python_post_3_6_only
def test_transpile_parameterize_excel_imports(tmp_path):
    tmp_file = str(tmp_path / 'txt.xlsx')
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df1.to_excel(tmp_file, index=False)

    mito = create_mito_wrapper()
    mito.excel_import(tmp_file, sheet_names=['Sheet1'], has_headers=True, skiprows=0)
    mito.excel_range_import(tmp_file, {'type': 'sheet name', 'value': 'Sheet1'}, [{'type': 'range', 'df_name': 'dataframe_1', 'value': 'A1:B2'}], convert_csv_to_xlsx=False)
    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'var_name': f"r'{tmp_file}'"}})

    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "import pandas as pd",
        "",
        "def function(var_name):",
        f"{TAB}sheet_df_dictonary = pd.read_excel(var_name, engine='openpyxl', sheet_name=[\n"
        f"{TAB*2}'Sheet1'\n"
        f"{TAB}], skiprows=0)",
        f"{TAB}Sheet1 = sheet_df_dictonary['Sheet1']",
        f'{TAB}',
        f"{TAB}dataframe_1 = pd.read_excel(var_name, sheet_name='Sheet1', skiprows=0, nrows=1, usecols='A:B')",
        f'{TAB}',
        f"{TAB}return Sheet1, dataframe_1",
        "",
        f"var_name = r'{tmp_file}'",
        "",
        f"Sheet1, dataframe_1 = function(var_name)"
    ]

def test_transpile_with_function_params_over_mitosheet():
    df1 = pd.DataFrame({'A': [1], 'B': [2]})
    df2 = pd.DataFrame({'A': [1], 'B': [2]})
    mito = create_mito_wrapper(df1, df2, arg_names=['df', 'df_copy'])
    mito.add_column(0, 'C')
    mito.add_column(1, 'C')

    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'param': "df"}})

    
    assert mito.transpiled_code == [
        'from mitosheet.public.v3 import *',
        "",
        "def function(param, df_copy):",
        f"{TAB}param.insert(2, 'C', 0)",
        f"{TAB}",
        f"{TAB}df_copy.insert(2, 'C', 0)",
        f"{TAB}",
        f"{TAB}return param, df_copy",
        "",
        f"param = df",
        "",
        f"param, df_copy = function(param, df_copy)"
    ]

def test_transpile_does_not_effect_chars_in_strings():
    mito = create_mito_wrapper()
    quote = '"'
    mito.ai_transformation(
        'do a test',
        'v1',
        'test',
        'test',
        """
df = pd.DataFrame({'A': ["has a new \
line in it", '\t', '     ']})
        """
    )

    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *
import pandas as pd

def function():
    df = pd.DataFrame({'A': ["has a new \
line in it", '\t', '     ']})
    
    return df

df = function()"""

def test_transpile_with_multiline_ai_completion():
    mito = create_mito_wrapper()
    mito.ai_transformation(
        'do a test',
        'v1',
        'test',
        'test',
        """
import pandas as pd

# create sample dataframe
df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6], 'C': [7, 8, 9]})

print(df)
        """
    )

    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

def function():
    import pandas as pd
    
    # create sample dataframe
    df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6], 'C': [7, 8, 9]})
    
    print(df)
    
    return df

df = function()"""

def test_transpiled_with_export_to_csv_singular():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, arg_names=['df'])
    mito.export_to_file('csv', [0], 'te"st.csv')

    assert [('df', 'df_name', 'Dataframe'), ("r'te" + '"' + "st.csv'", 'file_name', 'CSV export file path')] == get_parameterizable_params({}, mito.mito_backend.steps_manager)

    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'path': "r'te" + '"' + "st.csv'"}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

def function(df, path):
    df.to_csv(path, index=False)
    
    return df

path = r'te"st.csv'

df = function(df, path)"""

def test_transpiled_with_export_to_csv_multiple():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, df, arg_names=['df1', 'df2'])
    mito.export_to_file('csv', [0, 1], 'test.csv')

    assert [('df1', 'df_name', 'Dataframe'), ('df2', 'df_name', 'Dataframe'), ("r'test_0.csv'", 'file_name', 'CSV export file path'), ("r'test_1.csv'", 'file_name', 'CSV export file path')] == get_parameterizable_params({}, mito.mito_backend.steps_manager)

    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'path_0': "r'test_0.csv'", 'path_1': "r'test_1.csv'"}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *

def function(df1, df2, path_0, path_1):
    df1.to_csv(path_0, index=False)
    df2.to_csv(path_1, index=False)
    
    return df1, df2

path_0 = r'test_0.csv'
path_1 = r'test_1.csv'

df1, df2 = function(df1, df2, path_0, path_1)"""

def test_transpiled_with_export_to_xlsx_single():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, arg_names=['df'])
    mito.export_to_file('excel', [0], "te'st.xlsx")

    assert [('df', 'df_name', 'Dataframe'), ('r"te' + "'" + 'st.xlsx"', 'file_name', 'Excel export file path')] == get_parameterizable_params({}, mito.mito_backend.steps_manager)

    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'path': 'r"te' + "'" + 'st.xlsx"'}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *
import pandas as pd

def function(df, path):
    with pd.ExcelWriter(path, engine="openpyxl") as writer:
        df.to_excel(writer, sheet_name="df", index=False)
    
    return df

path = r"te'st.xlsx"

df = function(df, path)"""

def test_transpiled_with_export_to_xlsx_multiple():
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, df, arg_names=['df1', 'df2'])
    mito.export_to_file('excel', [0, 1], 'test.xlsx')

    assert [('df1', 'df_name', 'Dataframe'), ('df2', 'df_name', 'Dataframe'), ("r'test.xlsx'", 'file_name', 'Excel export file path')] == get_parameterizable_params({}, mito.mito_backend.steps_manager)

    mito.code_options_update({'as_function': True, 'call_function': True, 'function_name': 'function', 'function_params': {'path_0': "r'test.xlsx'"}})

    assert "\n".join(mito.transpiled_code) == """from mitosheet.public.v3 import *
import pandas as pd

def function(df1, df2, path_0):
    with pd.ExcelWriter(path_0, engine="openpyxl") as writer:
        df1.to_excel(writer, sheet_name="df1", index=False)
        df2.to_excel(writer, sheet_name="df2", index=False)
    
    return df1, df2

path_0 = r'test.xlsx'

df1, df2 = function(df1, df2, path_0)"""





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.export_to_file_code_chunk import ExportToFileCodeChunk
from mitosheet.excel_utils import get_df_name_as_valid_sheet_name
import pandas as pd
import os

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param

from mitosheet.utils import write_to_excel

from mitosheet.user.utils import is_pro, is_running_test

def get_export_to_csv_sheet_index_to_file_name(file_name: str, sheet_indexes: List[int]) -> Dict[int, str]:
    """
    Note that we add the index to the file name if there are multiple sheets being exported,
    and we do not add the df name. 

    This is because if you are parameterizing this code chunk, and then you change the imported
    data, the file name will change, which will break the parameterization.

    So we need to make the file name independent of the df name, so that changing the df name does
    not result in a broken parameterization.
    """
    if len(sheet_indexes) == 1:
        return {sheet_indexes[0]: f'{file_name}'}
    else:
        file_name_without_extension, extension = os.path.splitext(file_name)
        return {sheet_index: f'{file_name_without_extension}_{index}{extension}' for index, sheet_index in enumerate(sheet_indexes)}

def get_export_to_excel_sheet_index_to_sheet_name(state: State, file_name: str, sheet_indexes: List[int]) -> Dict[int, str]:
    return {sheet_index: get_df_name_as_valid_sheet_name(state.df_names[sheet_index]) for sheet_index in sheet_indexes}


def get_final_file_name(file_name: str, _type: str) -> str:
    _, extension = os.path.splitext(file_name)
    if extension == '':
        if _type == 'csv':
            return file_name + '.csv'
        elif _type == 'excel':
            return file_name + '.xlsx'
    
    return file_name


def upgrade_export_to_file_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    '''
    Adds a new param for exporting the formatting to excel which defaults to false. 
    '''
    params = step['params']
    params['export_formatting'] = False
    return [{
        "step_version": 2, 
        "step_type": "export_to_file", 
        "params": params
    }] + later_steps

class ExportToFileStepPerformer(StepPerformer):
    """
    Allows you to export to file.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'export_to_file'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        _type: str = get_param(params, 'type')
        sheet_indexes: List[int] = get_param(params, 'sheet_indexes')
        _file_name: str = get_param(params, 'file_name')
        _export_formatting: bool = get_param(params, 'export_formatting')

        file_name = get_final_file_name(_file_name, _type) # Ensure that the file name has the correct extension
        
        # We make a new state to modify it
        post_state = prev_state.copy()

        pandas_start_time = perf_counter()

        if _type == 'csv':
            sheet_index_to_export_location = get_export_to_csv_sheet_index_to_file_name(file_name, sheet_indexes)
            for sheet_index, file_name in sheet_index_to_export_location.items():
                post_state.dfs[sheet_index].to_csv(file_name, index=False)
        elif _type == 'excel':
            # Formatting is a Mito pro feature, but we also allow it for testing
            allow_formatting = (is_pro() or is_running_test()) and _export_formatting
            sheet_index_to_export_location = get_export_to_excel_sheet_index_to_sheet_name(post_state, file_name, sheet_indexes)
            write_to_excel(file_name, sheet_indexes, post_state, allow_formatting=allow_formatting)
        else:
            raise ValueError(f"Invalid file type: {_type}")

        pandas_processing_time = perf_counter() - pandas_start_time

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'sheet_index_to_export_location': sheet_index_to_export_location,
            'file_name': file_name,
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ExportToFileCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'type'),
                get_param(execution_data if execution_data is not None else {}, 'file_name'),
                get_param(execution_data if execution_data is not None else {}, 'sheet_index_to_export_location'),
                get_param(params, 'export_formatting')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return set()
    


from typing import Any, List

import pandas as pd

from mitosheet.types import ColumnHeader


def flatten_column_header(column_header: ColumnHeader) -> ColumnHeader:
    """
    Given a pandas column header, if it is a list or tuple, it will
    flatten this header into a string. 

    This is useful for headers that are created from a pivot table,
    where they are often tuples or lists.

    NOTE: This function, if applied before the depricated make_valid_header
    function, should return the same result for any input. That is:
    For all column headers:
    make_valid_header(flatten_column_header(ch)) = make_valid_header(ch)
    """
    if isinstance(column_header, list) or isinstance(column_header, tuple):
        column_header_str = [str(ch) for ch in column_header]
        return ' '.join(column_header_str).strip()

    return column_header

def deduplicate_column_headers(columns: List[ColumnHeader]) -> List[ColumnHeader]:
    """
    Ensures that all column headers are deduplicated, taking special care to
    handle NaNs.
    """

    final_column_headers = []
    seen_column_headers = set()
    nans_found = 0 # Special case, because it's not hashable
    for column_header in columns:
        final_column_header = column_header

        if isinstance(final_column_header, float) and pd.isna(final_column_header):
            if nans_found > 0:
                final_column_header = f'nan ({nans_found})'
                nans_found += 1
            else:
                nans_found += 1
        elif final_column_header in seen_column_headers:
            header_count = 1
            final_column_header = f'{column_header} ({header_count})'
            while final_column_header in seen_column_headers:
                header_count += 1
                final_column_header = f'{column_header} ({header_count})'

        final_column_headers.append(final_column_header)
        seen_column_headers.add(final_column_header)

    return final_column_headers




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that can be used in a sheet that operate on strings.

NOTE: This file is alphabetical order!
"""
from typing import Callable, Optional, Union
import sys
import numpy as np

import pandas as pd

from mitosheet.public.v3.errors import handle_sheet_function_errors
from mitosheet.public.v3.sheet_functions.utils import get_final_result_series_or_primitive, get_index_from_series, get_series_from_primitive_or_series
from mitosheet.public.v3.types.decorators import cast_values_in_all_args_to_type, cast_values_in_arg_to_type
from mitosheet.public.v3.types.sheet_function_types import StringInputType, StringRestrictedInputType, StringFunctionReturnType, IntRestrictedInputType, IntFunctionReturnType


@cast_values_in_arg_to_type('arg', 'str')
@handle_sheet_function_errors
def CLEAN(arg: StringRestrictedInputType) -> StringFunctionReturnType:
    """
    {
        "function": "CLEAN",
        "description": "Returns the text with the non-printable ASCII characters removed.",
        "search_terms": ["clean", "trim", "remove"],
        "examples": [
            "CLEAN(A)"
        ],
        "syntax": "CLEAN(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series whose non-printable characters are to be removed."
            }
        ]
    }
    """
    clean_helper: Callable[[str], str] = lambda x:''.join([i if 32 <= ord(i) < 126 else "" for i in x])

    if isinstance(arg, str):
        return clean_helper(arg)
    
    return arg.fillna('').apply(clean_helper)


@cast_values_in_all_args_to_type('str')
@handle_sheet_function_errors
def CONCAT(*argv: StringInputType) -> StringFunctionReturnType:
    """
    {
        "function": "CONCAT",
        "description": "Returns the passed strings and series appended together.",
        "search_terms": ["&", "concatenate", "append", "combine"],
        "examples": [
            "CONCAT('Bite', 'the bullet')",
            "CONCAT(A, B)"
        ],
        "syntax": "CONCAT(string1, [string2, ...])",
        "syntax_elements": [{
                "element": "string1",
                "description": "The first string or series."
            },
            {
                "element": "string2, ... [OPTIONAL]",
                "description": "Additional strings or series to append in sequence."
            }
        ]
    }
    """

    return get_final_result_series_or_primitive(
        '',
        argv,
        lambda df: df.sum().sum(), # Concats in the same order as Excel - yay!
        lambda previous_value, new_value: previous_value + new_value,
        lambda previous_series, new_series: previous_series + new_series
    )

@cast_values_in_arg_to_type('string', 'str')
@cast_values_in_arg_to_type('substrings', 'str')
@handle_sheet_function_errors
def FIND(string: StringRestrictedInputType, substrings: StringRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "FIND",
        "description": "Returns the position at which a string is first found within text, case-sensitive. Returns 0 if not found.",
        "search_terms": ["find", "search"],
        "examples": [
            "FIND(A, 'Jack')",
            "FIND('Ben has a friend Jack', 'Jack')"
        ],
        "syntax": "FIND(text_to_search, search_for)",
        "syntax_elements": [{
                "element": "text_to_search",
                "description": "The text or series to search for the first occurrence of search_for."
            },
            {
                "element": "search_for",
                "description": "The string to look for within text_to_search."
            }
        ]
    }
    """

    if isinstance(string, str) and isinstance(substrings, str):
        if substrings == '':
            return 1
        
        return string.find(substrings) + 1

    # otherwise, turn them into series
    index = get_index_from_series(string, substrings)
    string = get_series_from_primitive_or_series(string, index).fillna('')
    substrings = get_series_from_primitive_or_series(substrings, index).fillna('XTREME_MITOSHEET_NULL_VALUE_1234567890_0987654321_NULL_VALUE_MITOSHEET_XTREME') # This is a hack to make sure that the find function returns 0 for null values
    
    return pd.Series(
        [s.find(ss) + 1 for s, ss in zip(string, substrings)],
        index=index
    )


@cast_values_in_arg_to_type('string', 'str')
@cast_values_in_arg_to_type('num_chars', 'int')
@handle_sheet_function_errors
def LEFT(string: StringRestrictedInputType, num_chars: Optional[IntRestrictedInputType]=None) -> StringFunctionReturnType:
    """
    {
        "function": "LEFT",
        "description": "Returns a substring from the beginning of a specified string.",
        "search_terms": ["left"],
        "examples": [
            "LEFT(A, 2)",
            "LEFT('The first character!')"
        ],
        "syntax": "LEFT(string, [number_of_characters])",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series from which the left portion will be returned."
            },
            {
                "element": "number_of_characters [OPTIONAL, 1 by default]",
                "description": "The number of characters to return from the start of string."
            }
        ]
    }
    """

    left_helper: Callable[[str, int], str] = lambda s, nc: s[:min(nc, len(s))]

    if num_chars is None:
        num_chars = 1

    if isinstance(string, str) and isinstance(num_chars, int):
        return left_helper(string, num_chars)

    # otherwise, turn them into series for simplicity
    index = get_index_from_series(string, num_chars)
    string = get_series_from_primitive_or_series(string, index).fillna('')
    num_chars = get_series_from_primitive_or_series(num_chars, index).fillna(0)

    return pd.Series(
        [left_helper(s, nc) for s, nc in zip(string, num_chars)],
        index=string.index
    )



@cast_values_in_arg_to_type('arg', 'str')
@handle_sheet_function_errors
def LEN(arg: StringRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "LEN",
        "description": "Returns the length of a string.",
        "search_terms": ["length", "size"],
        "examples": [
            "LEN(A)",
            "LEN('This is 21 characters')"
        ],
        "syntax": "LEN(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series whose length will be returned."
            }
        ]
    }
    """
    if isinstance(arg, str):
        return len(arg)
    
    return arg.fillna('').str.len()


@cast_values_in_arg_to_type('arg', 'str')
@handle_sheet_function_errors
def LOWER(arg: StringRestrictedInputType) -> StringFunctionReturnType:
    """
    {
        "function": "LOWER",
        "description": "Converts a given string to lowercase.",
        "search_terms": ["lowercase", "uppercase"],
        "examples": [
            "=LOWER('ABC')",
            "=LOWER(A)",
            "=LOWER('Nate Rush')"
        ],
        "syntax": "LOWER(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to convert to lowercase."
            }
        ]
    }
    """
    if isinstance(arg, str):
        return arg.lower()

    return arg.fillna('').str.lower()



@cast_values_in_arg_to_type('string', 'str')
@cast_values_in_arg_to_type('start_loc', 'int')
@cast_values_in_arg_to_type('num_chars', 'int')
@handle_sheet_function_errors
def MID(string: StringRestrictedInputType, start_loc: IntRestrictedInputType, num_chars: IntRestrictedInputType) -> StringFunctionReturnType:
    """
    {
        "function": "MID",
        "description": "Returns a segment of a string.",
        "search_terms": ["middle"],
        "examples": [
            "MID(A, 2, 2)",
            "MID('Some middle characters!', 3, 4)"
        ],
        "syntax": "MID(string, starting_at, extract_length)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to extract the segment from."
            },
            {
                "element": "starting_at",
                "description": "The index from the left of string from which to begin extracting."
            },
            {
                "element": "extract_length",
                "description": "The length of the segment to extract."
            }
        ]
    }
    """

    if isinstance(string, str) and isinstance(start_loc, int) and isinstance(num_chars, int):
        return string[start_loc - 1:start_loc + num_chars - 1]

    # Turn all of them into a series to simplify things
    index = get_index_from_series(string, start_loc, num_chars)
    string = get_series_from_primitive_or_series(string, index).fillna('')
    start_loc = get_series_from_primitive_or_series(start_loc, index).fillna(0)
    num_chars = get_series_from_primitive_or_series(num_chars, index).fillna(0)
    
    return pd.Series(
        [s[sl-1:sl+nc-1] for s, sl, nc in zip(string, start_loc, num_chars)],
        index=index
    )


@cast_values_in_arg_to_type('arg', 'str')
@handle_sheet_function_errors
def PROPER(arg: StringRestrictedInputType) -> StringFunctionReturnType:
    """
    {
        "function": "PROPER",
        "description": "Capitalizes the first letter of each word in a specified string.",
        "search_terms": ["proper", "capitalize"],
        "examples": [
            "=PROPER('nate nush')",
            "=PROPER(A)"
        ],
        "syntax": "PROPER(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The value or series to convert to convert to proper case."
            }
        ]
    }
    """
    if isinstance(arg, str):
        return arg.title()

    return arg.fillna('').str.title()


@cast_values_in_arg_to_type('string', 'str')
@cast_values_in_arg_to_type('num_chars', 'int')
@handle_sheet_function_errors
def RIGHT(string: StringRestrictedInputType, num_chars: Optional[IntRestrictedInputType]=None) -> StringFunctionReturnType:

    """
    {
        "function": "RIGHT",
        "description": "Returns a substring from the beginning of a specified string.",
        "search_terms": [],
        "examples": [
            "RIGHT(A, 2)",
            "RIGHT('The last character!')"
        ],
        "syntax": "RIGHT(string, [number_of_characters])",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series from which the right portion will be returned."
            },
            {
                "element": "number_of_characters [OPTIONAL, 1 by default]",
                "description": "The number of characters to return from the end of string."
            }
        ]
    }
    """
    right_helper: Callable[[str, int], str] = lambda s, nc: s[-min(nc, len(s)):] if nc != 0 else ''

    if num_chars is None:
        num_chars = 1

    if isinstance(string, str) and isinstance(num_chars, int):
        return right_helper(string, num_chars)
    
    index = get_index_from_series(string, num_chars)
    string = get_series_from_primitive_or_series(string, index).fillna('')
    num_chars = get_series_from_primitive_or_series(num_chars, index).fillna(0)

    return pd.Series(
        [right_helper(s, nc) for s, nc in zip(string, num_chars)],
        index=string.index
    )


@cast_values_in_arg_to_type('string', 'str')
@cast_values_in_arg_to_type('old_text', 'str')
@cast_values_in_arg_to_type('new_text', 'str')
@cast_values_in_arg_to_type('count', 'int')
@handle_sheet_function_errors
def SUBSTITUTE(string: StringRestrictedInputType, old_text: StringRestrictedInputType, new_text: StringRestrictedInputType, count: Optional[IntRestrictedInputType]=None) -> StringFunctionReturnType:
    """
    {
        "function": "SUBSTITUTE",
        "description": "Replaces existing text with new text in a string.",
        "search_terms": ["replace", "find and replace"],
        "examples": [
            "SUBSTITUTE('Better great than never', 'great', 'late')",
            "SUBSTITUTE(A, 'dog', 'cat')"
        ],
        "syntax": "SUBSTITUTE(text_to_search, search_for, replace_with, [count])",
        "syntax_elements": [{
                "element": "text_to_search",
                "description": "The text within which to search and replace."
            },
            {
                "element": "search_for",
                "description": " The string to search for within text_to_search."
            },
            {
                "element": "replace_with",
                "description": "The string that will replace search_for."
            },
            {
                "element": "count",
                "description": "The number of times to perform the substitute. Default is all."
            }
        ]
    }
    """

    if count is None:
        count = -1

    if isinstance(string, str) and isinstance(old_text, str) and isinstance(new_text, str) and isinstance(count, int):
        return string.replace(old_text, new_text, count)

    index = get_index_from_series(string, old_text, new_text, count)
    string = get_series_from_primitive_or_series(string, index).fillna('')
    old_text = get_series_from_primitive_or_series(old_text, index).fillna('')
    new_text = get_series_from_primitive_or_series(new_text, index).fillna('')
    count = get_series_from_primitive_or_series(count, index).fillna(0)

    return pd.Series(
        [s.replace(ot, nt, c) for s, ot, nt, c in zip(string, old_text, new_text, count)],
        index=index
    )


@cast_values_in_arg_to_type('arg', 'str')
@handle_sheet_function_errors
def TEXT(arg: StringRestrictedInputType) -> StringFunctionReturnType:
    """
    {
        "function": "TEXT",
        "description": "Turns the passed series into a string.",
        "search_terms": ["string", "dtype"],
        "examples": [
            "=TEXT(Product_Number)",
            "=TEXT(Start_Date)"
        ],
        "syntax": "TEXT(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to convert to a string."
            }
        ]
    }
    """
    if isinstance(arg, str):
        return arg
    
    return arg.fillna('')

@cast_values_in_arg_to_type('arg', 'str')
@handle_sheet_function_errors
def TRIM(arg: StringRestrictedInputType) -> StringFunctionReturnType:
    """
    {
        "function": "TRIM",
        "description": "Returns a string with the leading and trailing whitespace removed.",
        "search_terms": ["trim", "whitespace", "spaces"],
        "examples": [
            "=TRIM('  ABC')",
            "=TRIM('  ABC  ')",
            "=TRIM(A)"
        ],
        "syntax": "TRIM(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The value or series to remove the leading and trailing whitespace from."
            }
        ]
    }
    """
    if isinstance(arg, str):
        return arg.strip()

    return arg.fillna('').str.strip()


@cast_values_in_arg_to_type('arg', 'str')
@handle_sheet_function_errors
def UPPER(arg: StringRestrictedInputType) -> StringFunctionReturnType:
    """
    {
        "function": "UPPER",
        "description": "Converts a given string to uppercase.",
        "search_terms": ["uppercase", "capitalize"],
        "examples": [
            "=UPPER('abc')",
            "=UPPER(A)",
            "=UPPER('Nate Rush')"
        ],
        "syntax": "UPPER(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to convert to uppercase."
            }
        ]
    }
    """
    if isinstance(arg, str):
        return arg.upper()
    
    return arg.fillna('').str.upper()



# TODO: we should see if we can list these automatically!
STRING_FUNCTIONS = {
    'CLEAN': CLEAN,
    'CONCAT': CONCAT,
    'FIND': FIND,
    'LEFT': LEFT,
    'LEN': LEN,
    'LOWER': LOWER,
    'MID': MID,
    'PROPER': PROPER,
    'RIGHT': RIGHT,
    'SUBSTITUTE': SUBSTITUTE,
    'TEXT': TEXT,
    'TRIM': TRIM,
    'UPPER': UPPER,
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State


class ConcatCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, join: str, ignore_index: bool, sheet_indexes: List[int]):
        super().__init__(prev_state, post_state)
        self.join = join
        self.ignore_index = ignore_index
        self.sheet_indexes = sheet_indexes

        self.df_names_to_concat = [self.post_state.df_names[sheet_index] for sheet_index in self.sheet_indexes]
        self.new_df_name = self.post_state.df_names[len(self.post_state.df_names) - 1]

    def get_display_name(self) -> str:
        return 'Concatenated'
    
    def get_description_comment(self) -> str:
        return f'Concatenated {len(self.sheet_indexes)} into dataframes into {self.new_df_name}'

    def get_code(self) -> Tuple[List[str], List[str]]:

        if len(self.df_names_to_concat) == 0:
            return [f'{self.new_df_name} = pd.DataFrame()'], ['import pandas as pd']
        else:
            return [
                f"{self.new_df_name} = pd.concat([{', '.join(self.df_names_to_concat)}], join=\'{self.join}\', ignore_index={self.ignore_index})"
            ], ['import pandas as pd']

    def get_created_sheet_indexes(self) -> List[int]:
        return [len(self.post_state.dfs) - 1]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
This file contains utilities for anonyimizing data. See the README.md in
this folder for more details on our approach to private telemetry.
"""

from typing import Any, Dict, Optional
from mitosheet.parser import parse_formula
from mitosheet.telemetry.private_params_map import LOG_PARAMS_FORMULAS, LOG_PARAMS_MAP_KEYS_TO_MAKE_PRIVATE, LOG_PARAMS_TO_LINEARIZE, LOG_PARAMS_PUBLIC
from mitosheet.types import StepsManagerType
from mitosheet.user.db import get_user_field
from mitosheet.user.schemas import UJ_USER_SALT

# When we anonymize, we use some combination of these words
# to construct new private words
valid_words = ['cat', 'dog', 'hat', 'time', 'person', 'year', 'way', 'thing', 'man', 'world', 'life', 'born', 'part', 'child', 'eye', 'woman', 'place', 'work', 'fall', 'case', 'point', 'company', 'number', 'group', 'problem', 'fact']

# We use the same salt to anonymize_words, and we read
# this salt in once the function is called for the first
# time, to make sure it's initialized properly
salt = None
def anonymize_as_string(word: Any) -> str:
    """
    Helper function that turns any specific value into
    a totally anonymous version of the value, consistently.

    Notably, this will cast the given value to a string before 
    anonymizing it.
    """
    # We make sure that the salt is read in after the entire
    # app has been initalized, so that we don't have to read
    # from the file all the time
    global salt
    if salt is None:
        salt = get_user_field(UJ_USER_SALT)

    word = str(word)

    # We select three indexes from the valid_words list, and concatenate them
    index_one = int(hash(salt + word + '0')) % len(valid_words)
    index_two = int(hash(salt + word + '1')) % len(valid_words)
    index_three = int(hash(salt + word + '2')) % len(valid_words)

    return valid_words[index_one] + valid_words[index_two] + valid_words[index_three]


def anonymize_formula(formula: str, sheet_index: int, steps_manager: Optional[StepsManagerType]=None) -> str:
    """
    Helper function that anonymizes formula to 
    make sure that no private data is included in it.
    """
    if steps_manager is None:
        return anonymize_as_string(formula)

    # We just input a random address, as we don't use it
    _, _, column_header_depedencies, index_label_dependencies = parse_formula(
        formula, 
        'A', 
        '0',
        {'type': 'entire_column'},
        steps_manager.dfs[sheet_index],
        throw_errors=False
    )
    
    for dependency in column_header_depedencies:
        formula = formula.replace(str(dependency), anonymize_as_string(dependency))
    for index_label in index_label_dependencies:
        formula = formula.replace(str(index_label), f"I[{anonymize_as_string(index_label)}]") # Tag with an I so we know it's an index
    
    return formula

def anonymize_object(obj: Any, anonymize_key: bool=False) -> Any:
    """
    Anoymizes any object it is given, handling any different
    type of object that it might be given.

    If obj is a dict and you want to anonymize the key instead of the value, 
    set anonymize_key=True
    """
    if isinstance(obj, list):
        return [anonymize_as_string(v) for v in obj]
    elif isinstance(obj, dict):
        if anonymize_key:
            return {anonymize_as_string(key) if anonymize_key else key: v for key, v in obj.items()}
        else: 
            return {key: anonymize_as_string(v) for key, v in obj.items()}
    
    return anonymize_as_string(obj)

def get_final_private_params_for_single_kv(key: str, value: Any, params: Dict[str, Any], steps_manager: Optional[StepsManagerType]=None) -> Dict[str, Any]:
    """
    Given a single key, value pair for a set of params, this function will 
    turn them into a totally anonyimized version of the parameter. 
    """
    private_params: Dict[str, Any] = dict()

    # If this is a log to linearize, then we recurse and create private versions of those
    # nested parameters. Note we only do this if it's actually possible to recurse
    if key in LOG_PARAMS_TO_LINEARIZE and isinstance(value, dict):
        for nested_key, nested_value in value.items():
            nested_params = get_final_private_params_for_single_kv(nested_key, nested_value, params, steps_manager)
            private_params = {
                **private_params, 
                # NOTE: we linearize the nested keys with {higher_key}_{lower_key} as the new key
                **{key + "_" + k: v for k, v in  nested_params.items()}
            }
    
        return private_params

    if key in LOG_PARAMS_PUBLIC:
        private_params[key] = value
    elif key in LOG_PARAMS_FORMULAS:
        private_params[key] = anonymize_formula(value, params['sheet_index'], steps_manager)
    else:
        private_params[key] = anonymize_object(value, anonymize_key=key in LOG_PARAMS_MAP_KEYS_TO_MAKE_PRIVATE)

    return private_params


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, List, Dict
from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import DEFAULT_DECIMAL, DEFAULT_DELIMETER, DEFAULT_ENCODING, DEFAULT_ERROR_BAD_LINES, DEFAULT_SKIPROWS
from mitosheet.data_in_mito import DataTypeInMito
from mitosheet.enterprise.mito_config import DEFAULT_MITO_CONFIG_SUPPORT_EMAIL, MEC_VERSION_KEYS, MitoConfig
from mitosheet.state import (
    DATAFRAME_SOURCE_DUPLICATED,
    DATAFRAME_SOURCE_IMPORTED,
    DATAFRAME_SOURCE_MERGED,
    DATAFRAME_SOURCE_PASSED,
    DATAFRAME_SOURCE_PIVOTED,
    DATAFRAME_SOURCE_TRANSPOSED,
    DATAFRAME_SOURCE_MELTED,
    NUMBER_FORMAT_ACCOUNTING,
    NUMBER_FORMAT_CURRENCY,
    NUMBER_FORMAT_PERCENTAGE,
    NUMBER_FORMAT_PLAIN_TEXT,
    NUMBER_FORMAT_SCIENTIFIC_NOTATION,
)
from mitosheet.step_performers import (
    STEP_PERFORMERS,
)
from mitosheet.step_performers.graph_steps.plotly_express_graphs import GRAPH_SAFETY_FILTER_CUTOFF
from mitosheet.step_performers.pivot import PIVOT_AGGREGATION_TYPES
from mitosheet.step_performers.sort import SORT_DIRECTION_ASCENDING, SORT_DIRECTION_DESCENDING
from mitosheet.user.schemas import USER_JSON_DEFAULT


def get_enum_from_ts_file(path, enum_name):
    """
    Helper function that retuns an enum from a TS
    file, parsed as a JSON object.

    NOTE: this only works for enums that are spread
    out across mulitple lines :-)
    """
    with open(path, "r+") as f:
        all_code_lines = f.readlines()

    # First, we find where the enum starts and ends
    start_line_index = None
    end_line_index = None
    for index, code_line in enumerate(all_code_lines):
        if f"enum {enum_name}" in code_line:
            start_line_index = index
        elif start_line_index is not None and "}" in code_line:
            end_line_index = index
            break

    if start_line_index is None:
        raise Exception(f"{enum_name} is not in {path}, did you move it or rename it?")

    # Then, for each item in the enum, we split it into key value pairs
    enum_item_code_lines = all_code_lines[start_line_index + 1 : end_line_index]
    enum_items = {}
    for enum_item_code_line in enum_item_code_lines:
        enum_item_code_line = enum_item_code_line.strip()
        # Remove trailing commas
        if enum_item_code_line.endswith(","):
            enum_item_code_line = enum_item_code_line[:-1]
        
        # If it's a comment, ignore it
        if enum_item_code_line.startswith('//'):
            continue

        # Split into keys and values
        [name, value] = enum_item_code_line.split(" = ")
        # If the value is a string and has quotes around it,
        # take them off
        if value.startswith("'") or value.startswith('"'):
            value = value[1:]
        if value.endswith("'") or value.endswith('"'):
            value = value[:-1]

        enum_items[name] = value

    return enum_items


def get_string_list_from_type(path, type_name):
    """
    Helper function that retuns an type from a TS
    file, parsed as a list of strings.

    NOTE: this only works for types that are spread
    out across mulitple lines :-)
    """
    with open(path, "r+") as f:
        all_code_lines = f.readlines()

    # First, we find where the enum starts and ends
    start_line_index = None
    end_line_index = None
    for index, code_line in enumerate(all_code_lines):
        if f"type {type_name}" in code_line:
            start_line_index = index
        elif start_line_index is not None and "}" in code_line:
            end_line_index = index
            break

    if start_line_index is None:
        raise Exception(f"{type_name} is not in {path}, did you move it or rename it?")

    items = []
    for type_code_line in all_code_lines[start_line_index + 1:]:
        type_code_line = type_code_line.strip()
        # Remove starting |
        if type_code_line.startswith("|"):
            type_code_line = type_code_line[1:]
        elif type_code_line.startswith('//'):
            # If it's a comment, ignore it
            continue
        else:
            # If there isn't one, then quit the loop
            break
        
        value = type_code_line.strip()
        # If the value is a string and has quotes around it,
        # take them off
        if value.startswith("'") or value.startswith('"'):
            value = value[1:]
        if value.endswith("'") or value.endswith('"'):
            value = value[:-1]

        items.append(value)



    return items


def get_constant_from_ts_file(path, constant_name):
    """
    Helper function that retuns an constant from a TS
    file, parsed as a JSON object.
    """
    with open(path, "r+") as f:
        all_code_lines = f.readlines()

    # First, we find where the constant
    line_index = None
    for index, code_line in enumerate(all_code_lines):
        if constant_name in code_line:
            line_index = index
            break

    if line_index is None:
        raise Exception(
            f"{constant_name} is not in {path}, did you move it or rename it?"
        )

    # Then, for each item in the enum, we split it into key value pairs
    [_, value] = all_code_lines[line_index].split(" = ")

    # Clean up the value a bit
    value = value.strip()
    if value.endswith(";"):
        value = value[:-1]

    return value

def get_keys_recursive(dictionary: Dict[str, Any], _keys: List[str]) -> List[str]:
    keys = _keys
    for key, value in dictionary.items():
        keys.append(key)
        if type(value) is dict:
            return get_keys_recursive(value, keys)
    return keys


def test_date_type_in_mito_match():
    data_type_in_mito = get_enum_from_ts_file("./src/mito/types.tsx", "DataTypeInMito")
    assert len(data_type_in_mito) == 4

    assert data_type_in_mito["NONE"] == DataTypeInMito.NONE.value
    assert data_type_in_mito["PROVIDED"] == DataTypeInMito.PROVIDED.value
    assert data_type_in_mito["TUTORIAL"] == DataTypeInMito.TUTORIAL.value
    assert data_type_in_mito["PERSONAL"] == DataTypeInMito.PERSONAL.value


def test_sort_direction_match():
    sort_direcion = get_enum_from_ts_file(
        "./src/mito/components/taskpanes/ControlPanel/FilterAndSortTab/SortCard.tsx",
        "SortDirection",
    )
    # NOTE: the extra element here is no sort (on the frontend)
    assert len(sort_direcion) == 3

    assert sort_direcion["ASCENDING"] == SORT_DIRECTION_ASCENDING
    assert sort_direcion["DESCENDING"] == SORT_DIRECTION_DESCENDING


def test_filter_conditions_match():
    # TODO: update test!
    pass


def test_pivot_aggregation_functions_match():
    aggregation_types = get_enum_from_ts_file(
        "./src/mito/types.tsx", "AggregationType"
    )
    assert set(aggregation_types.values()) == set(PIVOT_AGGREGATION_TYPES)


def test_step_types_match():
    step_types = get_enum_from_ts_file("./src/mito/types.tsx", "StepType")
    assert set(step_types.values()) == set(
        [step_performers.step_type() for step_performers in STEP_PERFORMERS]
        + ["initialize"]
    )


def test_df_sources_match():
    df_sources = get_enum_from_ts_file("./src/mito/types.tsx", "DFSource")
    assert set(df_sources.values()) == set(
        [
            DATAFRAME_SOURCE_PASSED,
            DATAFRAME_SOURCE_IMPORTED,
            DATAFRAME_SOURCE_PIVOTED,
            DATAFRAME_SOURCE_MERGED,
            DATAFRAME_SOURCE_DUPLICATED,
            DATAFRAME_SOURCE_TRANSPOSED,
            DATAFRAME_SOURCE_MELTED
        ]
    )


def test_user_json_fields_match():
    user_json_fields = get_enum_from_ts_file("./src/mito/api/api.tsx", "UserJsonFields")
    assert set(user_json_fields.values()) == set(USER_JSON_DEFAULT.keys())


def test_format_types_fields_match():
    format_types = get_enum_from_ts_file("./src/mito/types.tsx", "NumberColumnFormatEnum")
    format_types_values = format_types.values()

    assert set(format_types_values) == set(
        [
            NUMBER_FORMAT_PLAIN_TEXT,
            NUMBER_FORMAT_PERCENTAGE,
            NUMBER_FORMAT_ACCOUNTING,
            NUMBER_FORMAT_CURRENCY,
            NUMBER_FORMAT_SCIENTIFIC_NOTATION,
        ]
    )


def test_graph_safety_filter_cutoff_matches():
    graph_safety_filter_cutoff = get_constant_from_ts_file(
        "./src/mito/components/taskpanes/Graph/GraphSetupTab.tsx",
        "GRAPH_SAFETY_FILTER_CUTOFF",
    )

    assert int(graph_safety_filter_cutoff) == GRAPH_SAFETY_FILTER_CUTOFF

def test_mito_enterprise_keys_match():
    mito_enterprise_config_keys = get_enum_from_ts_file("./src/mito/types.tsx", "MitoEnterpriseConfigKey")

    # Since the MitoConfig is not set, we'll get none of the nested keys in the 
    # CodeSnippets object, so we use the MitoConfig and the MEC_VERSION_KEYS to 
    # get everything we need to test! 
    mito_config = MitoConfig().get_mito_config()
    keys = get_keys_recursive(mito_config, [])
    keys = keys + MEC_VERSION_KEYS['2']
    assert set(mito_enterprise_config_keys.values()) == set(keys)

def test_user_profile_defaults_matches():
    user_profile_support_email = get_constant_from_ts_file(
        "./src/mito/components/elements/GetSupportButton.tsx",
        'DEFAULT_SUPPORT_EMAIL'
    )

    assert user_profile_support_email == f"\'{DEFAULT_MITO_CONFIG_SUPPORT_EMAIL}\'"

def test_update_events_enum_defined():
    update_types = get_enum_from_ts_file("./src/mito/types.tsx", "UpdateType")
    from mitosheet.updates import UPDATES

    for UPDATE in UPDATES:
        assert UPDATE['event_type'] in update_types.values()
        
def test_pivot_column_transformation_type_defined():
    pcts = get_string_list_from_type("./src/mito/types.tsx", "PivotColumnTransformation")
    from mitosheet.step_performers.pivot import (
        PCT_NO_OP,
        PCT_DATE_YEAR,
        PCT_DATE_QUARTER,
        PCT_DATE_MONTH,
        PCT_DATE_WEEK,
        PCT_DATE_DAY_OF_MONTH,
        PCT_DATE_DAY_OF_WEEK,
        PCT_DATE_HOUR,
        PCT_DATE_MINUTE,
        PCT_DATE_SECOND,
        PCT_DATE_YEAR_MONTH_DAY_HOUR_MINUTE,
        PCT_DATE_YEAR_MONTH_DAY_HOUR,
        PCT_DATE_YEAR_MONTH_DAY,
        PCT_DATE_YEAR_MONTH,
        PCT_DATE_YEAR_QUARTER,
        PCT_DATE_MONTH_DAY,
        PCT_DATE_DAY_HOUR,
        PCT_DATE_HOUR_MINUTE,
    )

    PCTS = [
        PCT_NO_OP,
        PCT_DATE_YEAR,
        PCT_DATE_QUARTER,
        PCT_DATE_MONTH,
        PCT_DATE_WEEK,
        PCT_DATE_DAY_OF_MONTH,
        PCT_DATE_DAY_OF_WEEK,
        PCT_DATE_HOUR,
        PCT_DATE_MINUTE,
        PCT_DATE_SECOND,
        PCT_DATE_YEAR_MONTH_DAY_HOUR_MINUTE,
        PCT_DATE_YEAR_MONTH_DAY_HOUR,
        PCT_DATE_YEAR_MONTH_DAY,
        PCT_DATE_YEAR_MONTH,
        PCT_DATE_YEAR_QUARTER,
        PCT_DATE_MONTH_DAY,
        PCT_DATE_DAY_HOUR,
        PCT_DATE_HOUR_MINUTE,
    ]

    assert PCTS == pcts


def test_update_events_default_import_decimal():
    default_delimiter = get_constant_from_ts_file("./src/mito/components/import/CSVImportConfigScreen.tsx", "DEFAULT_DELIMETER")
    default_encoding = get_constant_from_ts_file("./src/mito/components/import/CSVImportConfigScreen.tsx", "DEFAULT_ENCODING")
    _default_decimal = get_constant_from_ts_file("./src/mito/components/import/CSVImportConfigScreen.tsx", "DEFAULT_DECIMAL")
    default_skiprows = get_constant_from_ts_file("./src/mito/components/import/CSVImportConfigScreen.tsx", "DEFAULT_SKIPROWS")
    default_error_bad_lines = get_constant_from_ts_file("./src/mito/components/import/CSVImportConfigScreen.tsx", "DEFAULT_ERROR_BAD_LINES")

    # We must make sure the DEFAULT_DECIMAL reference to the Decimal enum is correct.
    decimalEnum = get_enum_from_ts_file("./src/mito/components/import/CSVImportConfigScreen.tsx", "Decimal")
    default_decimal = decimalEnum[_default_decimal.split('.')[1]]
    
    assert default_delimiter == f'"{DEFAULT_DELIMETER}"'
    assert default_encoding == f'"{DEFAULT_ENCODING}"'
    assert default_decimal == f'{DEFAULT_DECIMAL}'
    assert int(default_skiprows) == DEFAULT_SKIPROWS
    assert bool(default_error_bad_lines) == DEFAULT_ERROR_BAD_LINES




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import \
    column_header_to_transpiled_code
from mitosheet.types import ColumnID


class ReorderColumnCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_id: ColumnID, new_column_index: int):
        super().__init__(prev_state, post_state)
        self.sheet_index: int = sheet_index
        self.column_id: ColumnID = column_id
        self.new_column_index: int = new_column_index

        self.df_name = self.post_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Reordered column'
    
    def get_description_comment(self) -> str:
        column_header = self.post_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        return f'Reordered column {column_header}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        from mitosheet.step_performers.column_steps.reorder_column import get_valid_index

        column_header = self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)
        transpiled_column_header = column_header_to_transpiled_code(column_header)

        new_column_index = get_valid_index(self.prev_state.dfs, self.sheet_index, self.new_column_index)

        # Get columns in df
        columns_list_line = f'{self.df_name}_columns = [col for col in {self.df_name}.columns if col != {transpiled_column_header}]'

        # Insert column into correct location 
        insert_line = f'{self.df_name}_columns.insert({new_column_index}, {transpiled_column_header})'
        
        # Apply reorder line
        apply_reorder_line = f'{self.df_name} = {self.df_name}[{self.df_name}_columns]'

        return [columns_list_line, insert_line, apply_reorder_line], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]



import pytest
import pandas as pd
from mitosheet.saved_analyses.save_utils import write_analysis

from mitosheet.tests.test_utils import create_mito_wrapper


def test_can_pass_sheet_function():
    def ADD1(col):
        return col + 1 
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}), sheet_functions=[ADD1])

    mito.set_formula('=ADD1(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

def test_can_pass_multiple_sheet_functions():
    def ADD1(col):
        return col + 1 
    def ADD2(col):
        return col + 2
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}), sheet_functions=[ADD1, ADD2])

    mito.set_formula('=ADD2(ADD1(A0))', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]}))


def test_pass_sheet_function_then_replay_errors_if_not_passed_again():
    def ADD1(col):
        return col + 1 
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, sheet_functions=[ADD1])

    mito.set_formula('=ADD1(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

    analysis_name = mito.mito_backend.analysis_name
    write_analysis(mito.mito_backend.steps_manager)

    new_mito = create_mito_wrapper(df)
    new_mito.replay_analysis(analysis_name)

    assert new_mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3]}))

def test_pass_sheet_function_then_replay_works_if_passed_again():
    def ADD1(col):
        return col + 1 
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, sheet_functions=[ADD1])

    mito.set_formula('=ADD1(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

    analysis_name = mito.mito_backend.analysis_name
    write_analysis(mito.mito_backend.steps_manager)

    new_mito = create_mito_wrapper(df, sheet_functions=[ADD1])
    new_mito.replay_analysis(analysis_name)

    assert new_mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

def test_sheet_functions_with_non_caps_error():
    def lower(col):
        return col + 1 
    df = pd.DataFrame({'A': [1, 2, 3]})
    with pytest.raises(ValueError):
        mito = create_mito_wrapper(df, sheet_functions=[lower])


def test_pass_sheet_function_returns_list():
    def COLUMN_SUM(df):
        return [df.sum() for x in range(len(df))]

    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, sheet_functions=[COLUMN_SUM])

    mito.set_formula('=COLUMN_SUM(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [6, 6, 6]}))





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
import pandas as pd
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.change_column_dtype_code_chunk import ChangeColumnDtypeCodeChunk

from mitosheet.errors import make_invalid_column_type_change_error, raise_error_if_column_ids_do_not_exist
from mitosheet.is_type_utils import (is_bool_dtype,
                                                   is_datetime_dtype,
                                                   is_float_dtype,
                                                   is_int_dtype, is_number_dtype,
                                                   is_string_dtype,
                                                   is_timedelta_dtype)
from mitosheet.public.v1.sheet_functions.types.utils import get_to_datetime_params
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnID


class ChangeColumnDtypeStepPerformer(StepPerformer):
    """"
    A step that allows changing the dtype of a column to a different
    dtype.

    Currently, supports: 'bool', 'int', 'float', 'str', 'datetime', 'timedelta'
    """

    @classmethod
    def step_version(cls) -> int:
        return 4

    @classmethod
    def step_type(cls) -> str:
        return 'change_column_dtype'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any]) -> Dict[str, Any]:
        sheet_index: int = params['sheet_index']
        column_ids: List[ColumnID] = params['column_ids']

        # Save all the old dtypes
        old_dtypes = dict()
        for column_id in column_ids:
            column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)
            old_dtypes[column_id] = str(prev_state.dfs[sheet_index][column_header].dtype)

        params['old_dtypes'] = old_dtypes
        
        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_ids: List[ColumnID] = get_param(params, 'column_ids')
        old_dtypes: Dict[ColumnID, str] = get_param(params, 'old_dtypes')
        new_dtype: str = get_param(params, 'new_dtype')
        public_interface_version: str = get_param(params, 'public_interface_version')

        raise_error_if_column_ids_do_not_exist(
            'change column dtype',
            prev_state,
            sheet_index,
            column_ids
        )

        if public_interface_version == 1:
            from mitosheet.public.v1 import to_int_series, to_boolean_series, to_float_series, to_timedelta_series, get_datetime_format
        elif public_interface_version == 2:
            from mitosheet.public.v2 import to_int_series, to_boolean_series, to_float_series, to_timedelta_series, get_datetime_format
        elif public_interface_version == 3:
            from mitosheet.public.v3 import to_int_series, to_boolean_series, to_float_series, to_timedelta_series, get_datetime_format
        else:
            raise Exception(f'Please add support for public_interface_version={public_interface_version}')

        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])
        pandas_processing_time: float = 0

        changed_column_ids: List[ColumnID] = []
        to_datetime_params_map: Dict[ColumnID, Dict[str, Any]] = {}
        for column_id in column_ids:

            old_dtype = old_dtypes[column_id]
            column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)

            column: pd.Series = prev_state.dfs[sheet_index][column_header]
            new_column = column
            changed_this_column = True
            
            # How we handle the type conversion depends on what type it is
            try:
                pandas_start_time = perf_counter()
                if is_bool_dtype(old_dtype):
                    if is_bool_dtype(new_dtype):
                        changed_this_column = False
                    elif is_int_dtype(new_dtype):
                        new_column = new_column.astype('int')
                    elif is_float_dtype(new_dtype):
                        new_column = column.astype('float')
                    elif is_string_dtype(new_dtype):
                        new_column = column.astype('str')
                    elif is_datetime_dtype(new_dtype):
                        raise make_invalid_column_type_change_error(
                            column_header,
                            old_dtype,
                            new_dtype
                        )
                    elif is_timedelta_dtype(new_dtype):
                        raise make_invalid_column_type_change_error(
                            column_header,
                            old_dtype,
                            new_dtype
                        )
                if is_int_dtype(old_dtype):
                    if is_bool_dtype(new_dtype):
                        new_column = column.fillna(False).astype('bool')
                    elif is_int_dtype(new_dtype):
                        changed_this_column = False
                    elif is_float_dtype(new_dtype):
                        new_column = column.astype('float')
                    elif is_string_dtype(new_dtype):
                        new_column = column.astype('str')
                    elif is_datetime_dtype(new_dtype):
                        new_column = pd.to_datetime(
                            column, 
                            unit='s',
                            errors='coerce'
                        )
                    elif is_timedelta_dtype(new_dtype):
                        new_column = to_timedelta_series(column)
                elif is_float_dtype(old_dtype):
                    if is_bool_dtype(new_dtype):
                        new_column = column.fillna(False).astype('bool')
                    elif is_int_dtype(new_dtype):
                        new_column = column.fillna(0).astype('int')
                    elif is_float_dtype(new_dtype):
                        changed_this_column = False
                    elif is_string_dtype(new_dtype):
                        new_column = column.astype('str')
                    elif is_datetime_dtype(new_dtype):
                        new_column = pd.to_datetime(
                            column, 
                            unit='s',
                            errors='coerce'
                        )
                    elif is_timedelta_dtype(new_dtype):
                        new_column = to_timedelta_series(column)
                elif is_string_dtype(old_dtype):
                    if is_bool_dtype(new_dtype):
                        new_column = to_boolean_series(new_column)
                    elif is_int_dtype(new_dtype):
                        new_column = to_int_series(column)
                    elif is_float_dtype(new_dtype):
                        new_column = to_float_series(column)
                    elif is_string_dtype(new_dtype):
                        changed_this_column = False
                    elif is_datetime_dtype(new_dtype):
                        # We need to detect the correct params to pass to the to_datetime function, and save them for use in the code chunk
                        to_datetime_params = get_to_datetime_params(column)
                        to_datetime_params_map[column_id] = to_datetime_params
                        
                        new_column = pd.to_datetime(
                            column,
                            errors='coerce',
                            **to_datetime_params
                        )
                    elif is_timedelta_dtype(new_dtype):
                        new_column = to_timedelta_series(column)
                elif is_datetime_dtype(old_dtype):
                    if is_bool_dtype(new_dtype):
                        new_column = ~column.isnull()
                    elif is_int_dtype(new_dtype):
                        new_column = column.astype('int') / 10**9
                    elif is_float_dtype(new_dtype):
                        # For some reason, we have to do all the conversions at once
                        new_column = column.astype('int').astype('float') / 10**9
                    elif is_string_dtype(new_dtype):
                        # NOTE: this is the same conversion that we send to the frontend
                        new_column = column.dt.strftime('%Y-%m-%d %X')
                    elif is_datetime_dtype(new_dtype):
                        changed_this_column = False
                    elif is_timedelta_dtype(new_dtype):
                        raise make_invalid_column_type_change_error(
                            column_header,
                            old_dtype,
                            new_dtype
                        )
                elif is_timedelta_dtype(old_dtype):
                    if is_bool_dtype(new_dtype):
                        new_column = ~column.isnull()
                    elif is_int_dtype(new_dtype):
                        new_column = column.dt.total_seconds().astype('int')
                    elif is_float_dtype(new_dtype):
                        new_column = column.dt.total_seconds()
                    elif is_string_dtype(new_dtype):
                        new_column = column.astype('str')
                    elif is_datetime_dtype(new_dtype):
                        raise make_invalid_column_type_change_error(
                            column_header,
                            old_dtype,
                            new_dtype
                        )
                    elif is_timedelta_dtype(new_dtype):
                        changed_this_column = False

                # We update the column, as well as the type of the column
                post_state.dfs[sheet_index][column_header] = new_column
                pandas_processing_time += (perf_counter() - pandas_start_time)

                # Remember which columns we changed
                if changed_this_column:
                    changed_column_ids.append(column_id)

                # If we're changing away from a number column, then we remove the formatting on the column if it exists
                if not is_number_dtype(new_dtype) and column_id in post_state.df_formats[sheet_index]['columns']:
                    del post_state.df_formats[sheet_index]['columns'][column_id]
                    
            except:
                raise make_invalid_column_type_change_error(
                    column_header,
                    old_dtype,
                    new_dtype
                )

        execution_data = {
            'pandas_processing_time': pandas_processing_time,
            'changed_column_ids': changed_column_ids
        }

        if len(to_datetime_params_map) > 0:
            execution_data['to_datetime_params_map'] = to_datetime_params_map

        return post_state, execution_data
        

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ChangeColumnDtypeCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_ids'),
                get_param(params, 'old_dtypes'),
                get_param(params, 'new_dtype'),
                get_param(execution_data if execution_data is not None else {}, 'changed_column_ids'),
                execution_data.get('to_datetime_params_map', None) if execution_data is not None else None,
                get_param(params, 'public_interface_version')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
from copy import deepcopy
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.bulk_old_rename.bulk_old_rename_code_chunk import BulkOldRenameCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.bulk_old_rename.deprecated_utils import \
    get_header_renames
from mitosheet.step_performers.column_steps.rename_column import \
    rename_column_headers_in_state
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param


class BulkOldRenameStepPerformer(StepPerformer):
    """
    This step just renames certain dataframes to have column
    headers that are renamed in the way we used to rename them.

    This makes it possible for us to upgrade to step versions
    that no longer do these renames, namely the preprocess step,
    simple import, and pivoting!
    """

    @classmethod
    def step_version(cls) -> int:
        # NOTE: we should _never_ change this step!
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'bulk_old_rename'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'old_rename_only_use_this_in_testing'

    @classmethod
    def execute(cls,prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        """
        If move_to_deprecated_id_algorithm, then this step is being
        used at the start of a replayed analysis to stand in for
        the preprocessing step that used to exist. 

        Thus, in this case, we move _all_ ids from the new format to 
        the old format. This is pretty unplesant, but necessary to 
        make sure that users who pass dataframes into Mito directly
        do not have their analysis break when they upgrade.
        """
        move_to_deprecated_id_algorithm: bool = get_param(params, 'move_to_deprecated_id_algorithm') if get_param(params, 'move_to_deprecated_id_algorithm') else False 

        post_state = prev_state.copy(deep_sheet_indexes=list(range(len(prev_state.dfs))))

        column_header_renames_list = []
        for sheet_index, df in enumerate(prev_state.dfs):
            column_headers = df.keys()
            column_header_renames = get_header_renames(column_headers)
            column_header_renames_list.append(column_header_renames)

            for old_column_header, new_column_header in column_header_renames.items():
                column_id = post_state.column_ids.get_column_id_by_header(sheet_index, old_column_header)
                rename_column_headers_in_state(
                    post_state,
                    sheet_index,
                    column_id,
                    new_column_header,
                    None
                )

        if move_to_deprecated_id_algorithm:
            post_state.move_to_deprecated_id_algorithm()
        
        return post_state, {
            'column_header_renames_list': column_header_renames_list
        }
        

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            BulkOldRenameCodeChunk(
                prev_state, 
                post_state, 
                execution_data.get('column_header_renames_list', []) if execution_data is not None else []
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return set() # changes all dataframes